from dataclasses import dataclass
from datetime import datetime, timedelta

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import from_tree
from ..client.cache import (
    Cache,
    CacheIsFull,
    NoCache,
    TooLargeForCache,
    WhenFull,
    download,
)
from ..client.utils import logger
from ..queries import FullText
from ..query_registration import register

tree = MapAdapter(
    {"arr": ArrayAdapter.from_array(numpy.arange(10), metadata={"a": 1})},
    metadata={"t": 1},
)


@pytest.fixture(params=["in_memory", "on_disk"])
def cache(request, tmpdir):
    if request.param == "in_memory":
        return Cache.in_memory(2e9)
    if request.param == "on_disk":
        return Cache.on_disk(tmpdir, capacity=2e9)


@pytest.mark.parametrize("structure_clients", ["numpy", "dask"])
def test_offline(cache, structure_clients):
    client = from_tree(tree, cache=cache, structure_clients=structure_clients)
    expected_tree_md = client.metadata
    expected_arr = client["arr"][:]
    if structure_clients == "dask":
        expected_arr = expected_arr.compute()
    expected_arr_md = client["arr"].metadata
    client["arr"]  # should be a cache hit

    # Switch this client into offline mode.
    client.offline = True
    actual_tree_md = client.metadata
    actual_arr = client["arr"][:]
    if structure_clients == "dask":
        actual_arr = actual_arr.compute()
    actual_arr_md = client["arr"].metadata
    assert numpy.array_equal(actual_arr, expected_arr)
    assert expected_arr_md == actual_arr_md
    assert expected_tree_md == actual_tree_md

    # Make a fresh client in offline mode from the start.
    client = from_tree(
        tree, cache=cache, structure_clients=structure_clients, offline=True
    )
    actual_tree_md = client.metadata
    actual_arr = client["arr"][:]
    if structure_clients == "dask":
        actual_arr = actual_arr.compute()
    actual_arr_md = client["arr"].metadata
    assert numpy.array_equal(actual_arr, expected_arr)
    assert expected_arr_md == actual_arr_md
    assert expected_tree_md == actual_tree_md

    # Switch online.
    client.offline = False


@pytest.mark.parametrize("structure_clients", ["numpy", "dask"])
def test_download(cache, structure_clients):
    client = from_tree(tree, cache=cache, structure_clients=structure_clients)
    download(client)
    assert cache.when_full == WhenFull.ERROR
    client.offline = True
    # smoke test
    client.metadata
    arr = client["arr"][:]
    if structure_clients == "dask":
        arr.compute()
    client["arr"].metadata


@pytest.mark.parametrize("structure_clients", ["numpy", "dask"])
def test_entries_stale_at(caplog, cache, structure_clients):
    """
    Test that entries_stale_at causes us to rely on the cache for some time.
    """
    logger.setLevel("DEBUG")
    EXPIRES_AFTER = 1  # seconds
    mapping = {"a": ArrayAdapter.from_array(numpy.arange(10), metadata={"a": 1})}
    tree = MapAdapter(
        mapping,
        metadata={"t": 1},
        entries_stale_after=timedelta(seconds=EXPIRES_AFTER),
    )
    client = from_tree(tree, cache=cache, structure_clients=structure_clients)
    client["a"].download()
    # Entries are stored in cache.
    assert "Cache stored" in caplog.text
    caplog.clear()
    for _ in range(3):
        client["a"].download()
        # Entries are read from cache *without contacting server*
        assert "Cache read" in caplog.text
        assert "Cache stored" not in caplog.text
        assert "<- 304" not in caplog.text
        caplog.clear()
    client["a"].download()
    # By default, we do not refresh stale entries.
    assert "Cache read" in caplog.text
    assert "<- 304" not in caplog.text
    assert "Cache stored" not in caplog.text
    caplog.clear()
    client["a"].refresh()
    # Server is contacted to confirm cache is still valid.
    # The cache is still valid. Entries are read from cache.
    assert "<- 304" in caplog.text
    assert "Cache read" in caplog.text
    assert "Cache stored" not in caplog.text
    caplog.clear()
    # Change the entries...
    mapping["b"] = ArrayAdapter.from_array(2 * numpy.arange(10), metadata={"b": 2})
    client["b"].download()
    assert "Cache stored" in caplog.text


@pytest.mark.parametrize("structure_clients", ["numpy", "dask"])
def test_content_stale_at(caplog, cache, structure_clients):
    """
    Test that metadata_stale_at causes us to rely on the cache for some time.
    """
    logger.setLevel("DEBUG")
    mapping = {"a": ArrayAdapter.from_array(numpy.arange(10))}
    tree = MapAdapter(
        mapping,
        entries_stale_after=timedelta(seconds=1_000),
    )
    # Monkey-patch this attribute on for now because we aren't decided
    # how to best to expose this when initializing Adapters.
    mapping["a"].content_stale_at = datetime.utcnow() + timedelta(seconds=1_000)
    mapping["a"].metadata_stale_at = datetime.utcnow() + timedelta(seconds=1_000)
    client = from_tree(tree, cache=cache, structure_clients=structure_clients)
    assert client["a"].read().sum() == 45
    # Content is stored in cache.
    assert "Cache stored" in caplog.text
    assert "Cache read" not in caplog.text
    caplog.clear()
    for _ in range(3):
        assert client["a"].read().sum() == 45
        # Content is read from cache *without contacting server*
        assert "Cache read" in caplog.text
        assert "Cache stored" not in caplog.text
        assert "<- 304" not in caplog.text
        caplog.clear()
    # This refresh will have no effect because nothing is stale.
    client["a"].refresh()
    assert "Cache read" in caplog.text
    assert "Cache stored" not in caplog.text
    assert "<- 304" not in caplog.text
    caplog.clear()
    # A force-refresh will have an effect.
    client["a"].refresh(force=True)
    # Server is contacted to confirm cache is still valid.
    # The cache is still valid. Content is read from cache.
    assert "Cache read" in caplog.text
    assert "Cache stored" not in caplog.text
    assert "<- 304" in caplog.text
    caplog.clear()
    # Change the content...
    mapping["a"] = ArrayAdapter.from_array(2 * numpy.arange(10))
    client["a"].refresh(force=True)
    # Server is contacted to confirm cache is still valid ("renew" it).
    # The cache is NOT still valid. Content is read from server.
    # Content is stored in cache.
    assert "Cache stored" in caplog.text
    assert client["a"].read().sum() == 90


@pytest.mark.parametrize("structure_clients", ["numpy", "dask"])
def test_metadata_stale_at(caplog, cache, structure_clients):
    """
    Test that metadata_stale_at causes us to rely on the cache for some time.
    """
    logger.setLevel("DEBUG")
    metadata = {"a": 1}
    mapping = {"a": ArrayAdapter.from_array(numpy.arange(10), metadata=metadata)}
    tree = MapAdapter(
        mapping,
        entries_stale_after=timedelta(seconds=1_000),
    )
    # Monkey-patch this attribute on for now because we aren't decided
    # how to best to expose this when initializing Adapters.
    mapping["a"].metadata_stale_at = datetime.utcnow() + timedelta(seconds=1_000)
    mapping["a"].content_stale_at = datetime.utcnow() + timedelta(seconds=1_000)
    client = from_tree(tree, cache=cache, structure_clients=structure_clients)
    assert client["a"].metadata["a"] == 1
    # Metadata are stored in cache.
    assert "Cache stored" in caplog.text
    assert "Cache read" not in caplog.text
    caplog.clear()
    for _ in range(3):
        assert client["a"].metadata["a"] == 1
        # Metadata are read from cache *without contacting server*
        assert "Cache read" in caplog.text
        assert "Cache stored" not in caplog.text
        caplog.clear()
    assert client["a"].metadata["a"] == 1
    caplog.clear()
    # This refresh will have an effect because we haven't downloaded everything yet.
    client["a"].refresh()
    assert "Cache read" in caplog.text
    assert "Cache stored" in caplog.text
    assert "<- 304" not in caplog.text
    caplog.clear()
    # A force-refresh will also have an effect.
    client["a"].refresh(force=True)
    # Server is contacted to confirm cache is still valid.
    # The cache is still valid. Metadata are read from cache.
    assert "Cache read" in caplog.text
    assert "Cache stored" not in caplog.text
    caplog.clear()
    # Change the metadata...
    metadata["a"] = 2
    client.refresh(force=True)
    # Server is contacted to confirm cache is still valid.
    # The cache is NOT still valid. Metadata are read from server.
    # Metadata are stored in cache.
    assert client["a"].metadata["a"] == 2
    assert "Cache stored" in caplog.text


def test_download_with_no_cache():
    tree = MapAdapter({})
    client = from_tree(tree)
    with pytest.raises(NoCache):
        client.download()


@register("stable")
@dataclass
class StableQuery:
    "A dummy query whose results will not change"
    dummy: str


def test_must_revalidate(caplog, cache):
    """
    By default, search results set Cache-Control: must-revalidate.
    Queries can override this if they expect the results to be very stable.
    (The motivating application is Bluesky scan_id lookup --- not _technically_
    unique but generally expected to be stable and not worth revalidating.)
    """

    def stable_query_search(query, tree):
        result = tree.new_variation(must_revalidate=False)
        return result

    logger.setLevel("DEBUG")
    mapping = {
        "a": ArrayAdapter.from_array(numpy.arange(10), metadata={"color": "red"})
    }
    tree = MapAdapter(
        mapping,
        metadata={"t": 1},
        must_revalidate=True,
    )
    tree.register_query(StableQuery, stable_query_search)
    client = from_tree(tree, cache=cache)
    caplog.clear()
    list(client.search(FullText("red")))
    assert "must-revalidate" in caplog.text

    # The must-revalidate Cache-Control directive forces a refresh.
    caplog.clear()
    list(client.search(FullText("red")))
    assert "<- 304" in caplog.text
    assert "must-revalidate" in caplog.text
    # But the results have not changed.
    assert "<- 200" not in caplog.text

    # The StableQuery does not require revalidation.
    caplog.clear()
    list(client.search(StableQuery(dummy="stuff")))
    assert "must-revalidate" not in caplog.text

    caplog.clear()
    list(client.search(StableQuery(dummy="stuff")))
    assert "must-revalidate" not in caplog.text
    assert "<- 304" not in caplog.text
    assert "<- 200" not in caplog.text


def test_when_full(caplog):
    logger.setLevel("DEBUG")
    arr = numpy.random.random((1000, 1000))
    tree = MapAdapter(
        {
            "a": ArrayAdapter.from_array(arr),
            "b": ArrayAdapter.from_array(arr),
            "c": ArrayAdapter.from_array(arr),
        }
    )

    # error
    cache = Cache.in_memory(1.5 * arr.nbytes)
    cache.when_full = "error"
    c = from_tree(tree, cache=cache)
    c["a"][:]
    with pytest.raises(CacheIsFull):
        c["b"][:]

    # warn
    cache = Cache.in_memory(1.5 * arr.nbytes)
    cache.when_full = "warn"
    c = from_tree(tree, cache=cache)
    c["a"][:]
    with pytest.warns(UserWarning):
        c["b"][:]

    # evict
    caplog.clear()
    cache = Cache.in_memory(1.5 * arr.nbytes)
    assert cache.when_full == WhenFull.EVICT  # default
    c = from_tree(tree, cache=cache)
    c["a"][:]
    assert "stored (8_000_000 B" in caplog.text
    caplog.clear()
    c["b"][:]
    assert "stored (8_000_000 B" in caplog.text
    assert "evicted 8_000_000 B" in caplog.text


def test_too_large(caplog):
    logger.setLevel("DEBUG")
    arr = numpy.random.random((1000, 1000))
    tree = MapAdapter(
        {
            "a": ArrayAdapter.from_array(arr),
        }
    )

    # error
    cache = Cache.in_memory(0.5 * arr.nbytes)
    cache.when_full = "error"
    c = from_tree(tree, cache=cache)
    c["a"]
    with pytest.raises(TooLargeForCache):
        c["a"][:]

    # warn
    cache = Cache.in_memory(0.5 * arr.nbytes)
    cache.when_full = "warn"
    c = from_tree(tree, cache=cache)
    c["a"]
    with pytest.warns(UserWarning):
        c["a"][:]

    # evict
    cache = Cache.in_memory(0.5 * arr.nbytes)
    assert cache.when_full == WhenFull.EVICT  # default
    c = from_tree(tree, cache=cache)
    c["a"]
    c["a"][:]



"""
This module includes objects inspired by https://github.com/dask/cachey/

We opted for an independent implementation because reusing cachey would have required:

* An invasive subclass that could be a bit fragile
* And also composition in order to get the public API we want
* Carrying around some complexity/features that we do not use here

The original cachey license (which, like Tiled's, is 3-clause BSD) is included in
the same source directory as this module.
"""
import collections.abc
import enum
import functools
import hashlib
import threading
import time
import types
import warnings
from collections import defaultdict
from datetime import datetime, timedelta
from math import log
from pathlib import Path
from urllib.parse import quote_plus

from heapdict import heapdict
from httpx import Headers

if __debug__:
    from .utils import logger


class Revalidate(enum.Enum):
    FORCE = "force"
    IF_EXPIRED = "if_expired"
    IF_WE_MUST = "if_we_must"

    def __call__(member):
        if isinstance(member, str):
            member = member.lower()
        return super().__call__(member)


class WhenFull(enum.Enum):
    EVICT = "evict"
    ERROR = "error"
    WARN = "warn"

    def __call__(member):
        if isinstance(member, str):
            member = member.lower()
        return super().__call__(member)


class UrlItem(
    collections.namedtuple(
        "UrlItem",
        [
            "size",
            "media_type",
            "encoding",
            "etag",
            "must_revalidate",
            "expires",
        ],
    )
):
    """
    An item in the cache mapping URLs to ETags (with other Headers info)
    """

    @classmethod
    def from_headers(cls, headers):
        expires_str = headers.get("expires")
        if expires_str is not None:
            expires = datetime.strptime(expires_str, HTTP_EXPIRES_HEADER_FORMAT)
        else:
            expires = None
        return cls(
            size=int(headers["content-length"]),
            media_type=headers["content-type"],
            encoding=None,
            # Record encoding when we start writing compressed data.
            # We currently always write it uncompressed, so the encoding is None.
            # encoding=headers.get("content-encoding"),
            etag=headers["etag"],
            must_revalidate="must-revalidate" in headers.get("cache-control", ""),
            expires=expires,
        )

    @classmethod
    def from_text(cls, text):
        headers = Headers()
        for line in text.splitlines():
            k, v = line.split(": ", 1)
            headers[k] = v
        return cls.from_headers(headers)

    def to_text(self):
        headers = {
            "content-length": str(self.size),
            "content-type": self.media_type,
            "etag": self.etag,
        }
        if self.must_revalidate:
            headers["cache-control"] = "must-revalidate"
        if self.expires is not None:
            headers["expires"] = self.expires.strftime(HTTP_EXPIRES_HEADER_FORMAT)
        if self.encoding is not None:
            headers["content-encoding"] = self.encoding
        return "\n".join(f"{k}: {v}" for k, v in headers.items())


def download(*entries):
    """
    Download a local cache for Tiled so access is fast and can work offline.

    Parameters
    ----------
    *entries : Node(s) or structure client(s)

    Examples
    --------

    Connect a tree and download it in its entirety.

    >>> from tiled.client import from_uri
    >>> from tiled.client.cache import download, Cache
    >>> client = from_uri("http://...", cache=Cache.on_disk("path/to/directory"))
    >>> download(client)

    Alternatively ,this can be done from the commandline via the tiled CLI:

    $ tiled download "http://..." my_cache_direcotry

    Use the local copy for faster data access. Tiled will connect to server just
    to verify that the local copy is current, and only download data if there have
    been changes to the copy of the server.

    >>> from tiled.client import from_uri
    >>> from tiled.client.cache import Cache
    >>> client = from_uri("http://...")
    >>> client = from_uri("http://...", cache=Cache.on_disk("my_cache_directory"))

    If network is unavailable or very slow, rely on the local copy entirely. Tiled
    will not connect to the server. (Note that you still need to provide a URL,
    but it is only used to contruct the names of files in the local directory.)

    >>> from tiled.client import from_uri
    >>> from tiled.client.cache import Cache
    >>> client = from_uri("http://...", cache=Cache.on_disk("my_cache_directory"), offline=True)
    """
    for entry in entries:
        entry.download()
    # Protect explicitly downloaded data from being silently evicted.
    entry.context.cache.when_full = WhenFull.ERROR


# This is a silly time format, but it is the HTTP standard.
HTTP_EXPIRES_HEADER_FORMAT = "%a, %d %b %Y %H:%M:%S GMT"
UNIT_SECOND = timedelta(seconds=1)
ZERO_SECONDS = timedelta(seconds=0)


def _round_seconds(dt):
    return round(dt / UNIT_SECOND)


class Reservation:
    """
    This represents a reservation on a cached piece of content.

    The content will not be evicted from the cache or updated
    until the content is loaded using `load_content()` or released
    using `ensure_released()`.
    """

    def __init__(self, url, item, renew, lock, load_content):
        self.url = url
        self.item = item
        self._renew = renew
        self._lock = lock
        self._lock_held = True
        self._load_content = load_content
        lock.acquire()

    def load_content(self):
        "Return the content and release the reservation."
        start = time.perf_counter()
        content = self._load_content()
        duration = 1000 * (time.perf_counter() - start)  # units: ms
        self._lock.release()
        if __debug__:
            # Use _ for thousands separator in bytes.
            logger.debug(
                "Cache read (%s B in %.1f ms) %s",
                f"{len(content):_}",
                duration,
                self.url,
            )
        return content

    def is_stale(self):
        if self.item.expires is None:
            stale = True
        else:
            time_remaining = datetime.utcnow() - self.item.expires
            stale = time_remaining > ZERO_SECONDS
        return stale

    def renew(self, expires):
        self._renew(expires=expires)

    def ensure_released(self):
        "Release the reservation. This is idempotent."
        if self._lock_held:
            try:
                self._lock.release()
                # TODO Investigate why this is sometimes released twice.
            except (AttributeError, RuntimeError):
                pass
                self._lock_held = False


class Cache:
    """
    A client-side cache of data from the server.

    The __init__ is to be used internally and by authors of custom caches.
    See ``Cache.in_memory()`` and ``Cache.on_disk()`` for user-facing methods.

    This is used by the function ``tiled.client.utils.get_content_with_cache``.
    """

    @classmethod
    def in_memory(cls, capacity, *, scorer=None):
        """
        An in-memory cache of data from the server

        This is useful to ensure that data is not downloaded repeatedly
        unless it has been updated since the last download.

        Because it is in memory, it only applies to a given Python process,
        i.e. a given working session. See ``Cache.on_disk()`` for a
        cache that can be shared across process and persistent for future
        sessions.

        Parameters
        ----------
        capacity : integer
            e.g. 2e9 to use up to 2 GB of RAM
        scorer : Scorer
            Determines which items to evict from the cache when it grows full.
            See tiled.client.cache.Scorer for example.
        """
        return cls(
            capacity,
            url_to_headers_cache={},
            etag_to_content_cache={},
            global_lock=threading.Lock(),
            lock_factory=lambda etag: threading.Lock(),
            state=types.SimpleNamespace(),
            scorer=scorer,
        )

    @classmethod
    def on_disk(
        cls,
        path,
        capacity=None,
        *,
        scorer=None,
    ):
        """
        An on-disk cache of data from the server

        This is useful to ensure that data is not downloaded repeatedly
        unless it has been updated since the last download.

        This uses file-based locking to ensure consistency when the cache
        is shared by multiple processes.

        Parameters
        ----------
        path : Path or str
            A directory will be created at this path if it does not yet exist.
            It is safe to reuse an existing cache directory and to share a cache
            directory between multiple processes.
        capacity : integer, optional
            e.g. 2e9 to use up to 2 GB of disk space. If None, this will consume
            up to (X - 1 GB) where X is the free space remaining on the volume
            containing `path`.
        scorer : Scorer
            Determines which items to evict from the cache when it grows full.
            See tiled.client.cache.Scorer for example.
        """
        import locket

        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        if capacity is None:
            # By default, use (X - 1 GB) where X is the current free space
            # on the volume containing `path`.
            import shutil

            capacity = shutil.disk_usage(path).free - 1e9
        etag_to_content_cache = FileBasedCache(path / "etag_to_content_cache")
        instance = cls(
            capacity,
            url_to_headers_cache=FileBasedUrlCache(path / "url_to_headers_cache"),
            etag_to_content_cache=etag_to_content_cache,
            global_lock=locket.lock_file(path / "global.lock"),
            lock_factory=lambda etag: locket.lock_file(
                path / "etag_to_content_cache" / f"{etag}.lock"
            ),
            state=OnDiskState(path),
            scorer=scorer,
        )
        return instance

    def __init__(
        self,
        capacity,
        *,
        url_to_headers_cache,
        etag_to_content_cache,
        global_lock,
        lock_factory,
        state,
        scorer=None,
    ):
        """
        Parameters
        ----------

        capacity : int
            The number of bytes of data to keep in the cache
        url_to_headers_cache : MutableMapping
            Dict-like object to use for cache
        etag_to_content_cache : MutableMapping
            Dict-like object to use for cache
        global_lock : Lock
            A lock used for the url_to_headers_cache
        lock_factory : callable
            Expected signature: ``f(etag) -> Lock``
        state : object
            Namespace used for storing general-purpose cache state
        scorer: Scorer, optional
            A Scorer object that controls how we decide what to retire when space
            is low.
        """

        if scorer is None:
            scorer = Scorer(halflife=1000)
        self.scorer = scorer
        self._state = state
        if not hasattr(state, "when_full"):
            state.when_full = b"evict"
        self.capacity = capacity
        self.heap = heapdict()
        self.nbytes = dict()
        self.total_bytes = 0
        self.url_to_headers_cache = url_to_headers_cache
        self.etag_to_content_cache = etag_to_content_cache
        self.etag_refcount = defaultdict(lambda: 0)
        self.etag_lock = LockDict.from_lock_factory(lock_factory)
        self.url_to_headers_lock = global_lock
        # If the cache has data in it, initialize the internal caches.
        for etag in etag_to_content_cache:
            # This tells us the content size without actually reading in the data.
            nbytes = etag_to_content_cache.sizeof(etag)
            score = self.scorer.touch(etag, nbytes)
            self.heap[etag] = score
            self.nbytes[etag] = nbytes
            self.total_bytes += nbytes

    @property
    def when_full(self):
        """
        Controls what happens the cache is at capacity

        * EVICT - Make a cost/benefit judgement about what (if anything) to evict to make room.
        * ERROR - Raise an error.
        * WARN - Leave the cache content alone and warn that nothing new will be cached.
          This is not recommended.
        """
        return WhenFull(self._state.when_full.decode())

    @when_full.setter
    def when_full(self, value):
        self._state.when_full = WhenFull(value).value.encode()

    def renew(self, url, etag, expires):
        cache_key = tokenize_url(url)
        if expires is None:
            # Do not renew.
            return
        with self.url_to_headers_lock:
            item = self.url_to_headers_cache[cache_key]
            assert item.etag == etag
            # TO DO We end up going str -> datetime -> str here.
            # It may be worth adding a fast path.
            expires_dt = datetime.strptime(expires, HTTP_EXPIRES_HEADER_FORMAT)
            updated_item = item._replace(expires=expires_dt)
            self.url_to_headers_cache[cache_key] = updated_item
        if __debug__:
            logger.debug(
                "Cache renewed %s for %d secs.",
                url,
                _round_seconds(expires_dt - datetime.utcnow()),
            )

    def put(self, url, headers, content):
        cache_key = tokenize_url(url)
        cached = self.url_to_headers_cache.get(cache_key)
        if cached:
            previous_etag = cached.etag
            self.etag_refcount[previous_etag] -= 1
            if self.etag_refcount[previous_etag] == 0:
                # All URLs that referred to this content have since
                # changed their ETags, so we can forget about this content.
                self.retire(previous_etag)
        item = UrlItem.from_headers(headers)
        start = time.perf_counter()
        with self.url_to_headers_lock:
            self.url_to_headers_cache[cache_key] = item
        nbytes = self._put_content(item.etag, content)
        duration = 1000 * (time.perf_counter() - start)  # units: ms
        if __debug__:
            if nbytes:
                if item.expires is not None:
                    logger.debug(
                        "Cache stored (%s B in %.1f ms) %s. Stale in %d secs.",
                        f"{nbytes:_}",  # Use _ for thousands separator.
                        duration,
                        url,
                        _round_seconds(item.expires - datetime.utcnow()),
                    )
                else:
                    logger.debug(
                        "Cache stored (%s B in %.1f ms) %s. Immediately stale.",
                        f"{nbytes:_}",  # Use _ for thousands separator.
                        duration,
                        url,
                    )
            else:
                logger.debug(
                    "Cache delined to store %s given its cost/benefit score.", url
                )

    def _put_content(self, etag, content):
        nbytes = len(content)
        if nbytes > self.capacity:
            msg = (
                f"A single item of size {nbytes} is too large for the "
                f"cache of capacity {self.capacity}."
            )
            if self.when_full == WhenFull.ERROR:
                raise TooLargeForCache(msg)
            elif self.when_full == WhenFull.WARN:
                warnings.warn(msg)
            # If we are set to EVICT, just do not bother storing items too large for the cache.
            return
        if nbytes + self.total_bytes > self.capacity:
            if self.when_full == WhenFull.ERROR:
                raise CacheIsFull(
                    f"""All {self.capacity} bytes of the cache's capacity are used. Options:
    1. Set larger capacity (and if necessary a different storage volume with more room).
    2. Choose a smaller set of entries to cache.
    3. Allow the cache to evict items that do not fit by setting cache.when_full = "evict".
    4. Let the cache sit as it is, keeping the items it has but not adding any more,
       This is not recommended but it is doable by setting cache.when_full = "warn"."""
                )
            elif self.when_full == WhenFull.WARN:
                warnings.warn(
                    f"""All {self.capacity} bytes of the cache's capacity are used. Options:
    1. Set larger capacity (and if necessary a different storage volume with more room).
    2. Choose a smaller set of entries to cache.
    3. Allow the cache to evict items that do not fit by setting cache.when_full = "evict".
    4. Let the cache sit as it is, keeping the items it has but not adding any more,
       This is not recommended but it is doable by setting cache.when_full = "warn"."""
                )
                return
        # If we reach here, either the data fits or we are going to
        # make a cost/benefit assessment of whether to evict things to make room for it.
        score = self.scorer.touch(etag, nbytes)
        if (
            nbytes + self.total_bytes < self.capacity
            or not self.heap
            or score > self.heap.peekitem()[1]
        ):
            self.shrink(target=self.capacity - nbytes)
            self.etag_to_content_cache[etag] = content
            self.heap[etag] = score
            self.nbytes[etag] = nbytes
            self.total_bytes += nbytes
            return nbytes

    def get_reservation(self, url):
        # Hold the global lock.
        with self.url_to_headers_lock:
            cached = self.url_to_headers_cache.get(tokenize_url(url))
            if cached is None:
                # We have nothing for this URL.
                return None
            # Acquire a targeted lock, and then release and the global lock.
            lock = self.etag_lock[cached.etag]
        return Reservation(
            url,
            cached,
            functools.partial(self.renew, url, cached.etag),
            lock,
            functools.partial(self._get_content_for_etag, cached.etag),
        )

    def _get_content_for_etag(self, etag):
        try:
            content = self.etag_to_content_cache[etag]
            # Access this item increases its score.
            score = self.scorer.touch(etag, len(content))
            self.heap[etag] = score
            return content
        except KeyError:
            return None

    def retire(self, etag):
        """Retire/remove a etag from the cache

        See Also:
            shrink
        """
        lock = self.etag_lock[etag]
        with lock:
            self.etag_to_content_cache.pop(etag)
            self.total_bytes -= self.nbytes.pop(etag)
            self.etag_lock.pop(etag)

    def _shrink_one(self):
        if self.heap.heap:
            # Retire the lowest-priority item that isn't locked.
            for score, etag, _ in self.heap.heap:
                lock = self.etag_lock[etag]
                if lock.acquire(blocking=False):
                    try:
                        self.heap.pop(etag)
                        self.etag_to_content_cache.pop(etag)
                        nbytes = self.nbytes.pop(etag)
                        self.total_bytes -= nbytes
                        self.etag_lock.pop(etag)
                    finally:
                        lock.release()
                    break
            if __debug__:
                logger.debug("Cache evicted %s B with ETag %s", f"{nbytes:_}", etag)

    def resize(self, capacity):
        """Resize the cache.

        Will fit the cache into capacity by calling `shrink()`.
        """
        self.capacity = capacity
        self.shrink()

    def shrink(self, target=None):
        """Retire keys from the cache until we're under bytes budget

        See Also:
            retire
        """
        if target is None:
            target = self.capacity
        if self.total_bytes <= target:
            return

        while self.total_bytes > target:
            self._shrink_one()

    def clear(self):
        while self.etag_to_content_cache:
            self._shrink_one()


class Scorer:
    """
    Object to track scores of cache

    Prefers items that have the following properties:

    1.  Expensive to download (bytes)
    3.  Frequently used
    4.  Recently used

    This object tracks both stated costs of keys and a separate score related
    to how frequently/recently they have been accessed.  It uses these to to
    provide a score for the key used by the ``Cache`` object, which is the main
    usable object.

    Examples
    --------

    >>> s = Scorer(halflife=10)
    >>> s.touch('x', cost=2)  # score is similar to cost
    2
    >>> s.touch('x')  # scores increase on every touch
    4.138629436111989
    """

    def __init__(self, halflife):
        self.cost = dict()
        self.time = defaultdict(lambda: 0)

        self._base_multiplier = 1 + log(2) / float(halflife)
        self.tick = 1
        self._base = 1

    def touch(self, key, cost=None):
        """Update score for key
        Provide a cost the first time and optionally thereafter.
        """
        time = self._base

        if cost is not None:
            self.cost[key] = cost
            self.time[key] += self._base
            time = self.time[key]
        else:
            try:
                cost = self.cost[key]
                self.time[key] += self._base
                time = self.time[key]
            except KeyError:
                return

        self._base *= self._base_multiplier
        return cost * time


def tokenize_url(url):
    """
    >>> tokenize_url(httpx_URL)
    ("urlsafe_host_and_port", "md5_hash_of_the_rest")
    """
    url_as_tuple = url.raw
    address = url_as_tuple[1].decode()
    if url_as_tuple[2]:
        address += f":{url_as_tuple[2]}"
    return (quote_plus(address), hashlib.md5(b"".join(url_as_tuple[3:])).hexdigest())


class OnDiskState:
    def __init__(self, directory):
        self._directory = Path(directory)
        self._ttl_cache = {}

        if self._directory.exists():
            import packaging.version

            version_filepath = self._directory / "client_library_version"
            if version_filepath.is_file():
                with open(version_filepath, "rt") as file:
                    # Check to see if this cache has an internal layout that we
                    # understand.
                    version = packaging.version.parse(file.read())
                    if version <= packaging.version.parse("0.1.0a43"):
                        raise ValueError(
                            f"Cache directory {directory} was used by an older version of Tiled. "
                            "It cannot be used by this version. Choose a different directory, "
                            "or delete all the cached data in that directory. (In the future "
                            "Tiled caching will be backward compatible, but not during alpha.)"
                        )
        self._directory.mkdir(parents=True, exist_ok=True)
        # Record the version which may be useful to future code that needs
        # to deal with a change in internal format or layout.
        with open(self._directory / "client_library_version", "wt") as file:
            from tiled import __version__

            file.write(__version__)

    def __getattr__(self, key):
        try:
            deadline, value = self._ttl_cache[key]
            if deadline < time.monotonic():
                return value
        except KeyError:
            pass
        try:
            with open(self._directory / key, "rb") as file:
                value = file.read()
        except FileNotFoundError:
            raise AttributeError(key)
        TIMEOUT = 1  # second
        deadline = TIMEOUT + time.monotonic()
        self._ttl_cache[key] = deadline, value
        return value

    def __setattr__(self, key, value):
        if key.startswith("_"):
            return super().__setattr__(key, value)
        TIMEOUT = 1  # second
        deadline = TIMEOUT + time.monotonic()
        self._ttl_cache[key] = deadline, value
        with open(self._directory / key, "wb") as file:
            return file.write(value)


class FileBasedCache(collections.abc.MutableMapping):
    "Locking is handled in the other layer, by Cache."

    def __init__(self, directory, mode="w"):
        self._directory = Path(directory)
        self._directory.mkdir(parents=True, exist_ok=True)

    def __repr__(self):
        return repr(dict(self))

    @property
    def directory(self):
        return self._directory

    def sizeof(self, key):
        # This is vestigial and can likely be removed.
        path = Path(self._directory, *_normalize(key))
        return path.stat().st_size

    def __getitem__(self, key):
        path = Path(self._directory, *_normalize(key))
        if not path.is_file():
            raise KeyError(key)
        with open(path, "rb") as file:
            return file.read()

    def __setitem__(self, key, value):
        path = Path(self._directory, *_normalize(key))
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "wb") as file:
            file.write(value)

    def __delitem__(self, key):
        path = Path(self._directory, *_normalize(key))
        path.unlink()

    def __len__(self):
        return len(list(self._directory.iterdir()))

    def __iter__(self):
        for path in self._directory.iterdir():
            parts = path.relative_to(self._directory).parts
            if len(parts) == 1:
                # top-level metadata like "client_library_version"
                continue
            yield _unnormalize(parts)

    def __contains__(self, key):
        path = Path(self._directory, *_normalize(key))
        return path.is_file()


def _normalize(key):
    if isinstance(key, str):
        key = (key,)
    # To avoid an overly large directory (which leads to slow performance)
    # divide into subdirectories beginning with the first two characters of
    # the contents' name.
    return (*key[:-1], key[-1][:2], key[-1])


def _unnormalize(key):
    return (*key[:-2], key[-1])


class LockDict(dict):
    @classmethod
    def from_lock_factory(cls, lock_factory):
        instance = cls()
        instance._lock_factory = lock_factory
        return instance

    def __missing__(self, key):
        value = self._lock_factory(key)
        self[key] = value
        return value


class FileBasedUrlCache(FileBasedCache):
    def __getitem__(self, key):
        data = super().__getitem__(key)
        return UrlItem.from_text(data.decode())

    def __setitem__(self, key, value):
        data = value.to_text().encode()
        super().__setitem__(key, data)


class AlreadyTooLarge(Exception):
    pass


class CacheIsFull(Exception):
    pass


class TooLargeForCache(Exception):
    pass


class NoCache(Exception):
    pass


def verify_cache(cache):
    if cache is None:
        raise NoCache(
            """To download, the Tiled client needs to be configured
with a cache, such as:

from tiled.client import from_uri
from tiled.client.cache import Cache

c = from_uri("http://tiled-demo.blueskyproject.io/api", cache = Cache.on_disk("data"))

See https://blueskyproject.io/tiled/tutorials/caching.html for details."""
        )



import dask.array
import numpy
import pytest
import xarray
import xarray.testing

from ..adapters.mapping import MapAdapter
from ..adapters.xarray import DataArrayAdapter, DatasetAdapter, VariableAdapter
from ..client import from_tree
from ..client import xarray as xarray_client

array = numpy.random.random((10, 10))
EXPECTED = {
    "variable": xarray.Variable(
        data=dask.array.from_array(array),
        dims=["x", "y"],
        attrs={"thing": "stuff"},
    ),
    "data_array": xarray.DataArray(
        xarray.Variable(
            data=dask.array.from_array(array),
            dims=["x", "y"],
            attrs={"thing": "stuff"},
        ),
        coords={
            "x": dask.array.arange(len(array)),
            "y": 10 * dask.array.arange(len(array)),
        },
    ),
    "dataset": xarray.Dataset(
        {
            "image": xarray.DataArray(
                xarray.Variable(
                    data=dask.array.from_array(array),
                    dims=["x", "y"],
                    attrs={"thing": "stuff"},
                ),
                coords={
                    "x": dask.array.arange(len(array)),
                    "y": 10 * dask.array.arange(len(array)),
                },
            ),
            "z": xarray.DataArray(data=dask.array.ones((len(array),))),
        },
        coords={"time": numpy.arange(len(array))},
    ),
    "wide": xarray.Dataset(
        {f"column_{i:03}": xarray.DataArray(i * numpy.ones(10)) for i in range(500)}
    ),
}

tree = MapAdapter(
    {
        "variable": VariableAdapter(EXPECTED["variable"]),
        "data_array": DataArrayAdapter.from_data_array(EXPECTED["data_array"]),
        "dataset": DatasetAdapter(EXPECTED["dataset"]),
        "wide": DatasetAdapter(EXPECTED["wide"]),
    }
)


@pytest.mark.parametrize("key", list(tree))
def test_xarrays(key):
    client = from_tree(tree)
    expected = EXPECTED[key]
    actual = client[key].read()
    xarray.testing.assert_equal(actual, expected)


def test_dataset_column_access():
    client = from_tree(tree)
    expected_dataset = tree["dataset"].read()
    client_dataset = client["dataset"]
    for col in expected_dataset:
        actual = client_dataset[col]
        expected = expected_dataset[col]
        xarray.testing.assert_equal(actual, expected)


def test_dataset_coord_access():
    client = from_tree(tree)
    expected_dataset = tree["dataset"].read()
    client_dataset = client["dataset"]

    # Coordinate on data variable
    actual = client_dataset["image"].coords["x"]
    expected = expected_dataset["image"].coords["x"]
    xarray.testing.assert_equal(actual, expected)
    # Circular reference
    actual = client_dataset["image"].coords["x"].coords["x"].coords["x"]
    xarray.testing.assert_equal(actual, expected)

    # Coordinate on dataset
    actual = client_dataset.coords["time"].read()
    expected = expected_dataset.coords["time"]
    xarray.testing.assert_equal(actual, expected)
    # Circular reference
    actual = (
        client_dataset.coords["time"]
        .read()
        .coords["time"]
        .coords["time"]
        .coords["time"]
    )
    xarray.testing.assert_equal(actual, expected)


def test_nested_coords():
    # Example from
    # https://xarray.pydata.org/en/stable/user-guide/data-structures.html#creating-a-dataset
    temp = 15 + 8 * numpy.random.randn(2, 2, 3)
    precip = 10 * numpy.random.rand(2, 2, 3)
    lon = [[-99.83, -99.32], [-99.79, -99.23]]
    lat = [[42.25, 42.21], [42.63, 42.59]]

    ds = xarray.Dataset(
        {
            "temperature": (["x", "y", "time"], temp),
            "precipitation": (["x", "y", "time"], precip),
        },
        coords={
            "lon": (["x", "y"], lon),
            "lat": (["x", "y"], lat),
            "time": [1, 2, 3],
            "reference_time": [11, 12, 13],
        },
    )
    tree = MapAdapter({"ds": DatasetAdapter(ds)})
    client = from_tree(tree)
    expected_dataset = ds
    client_dataset = client["ds"].read()
    xarray.testing.assert_equal(client_dataset, expected_dataset)


def test_url_limit_handling():
    "Check that requests and split up to stay below the URL length limit."
    expected = tree["wide"].read()
    client = from_tree(tree)
    client["wide"].read()  # Dry run to run any one-off state-initializing requests.
    # Accumulate Requests here for later inspection.
    requests = []

    async def accumulate(request):
        # httpx.AsyncClient requires event hooks to be async functions.
        requests.append(request)

    client.context.event_hooks["request"].append(accumulate)
    actual = client["wide"].read()
    xarray.testing.assert_equal(actual, expected)
    normal_request_count = len(requests)
    original = xarray_client.URL_CHARACTER_LIMIT
    try:
        # It should never be necessary to tune this for real-world use, but we
        # use this knob as a way to test its operation.
        xarray_client.URL_CHARACTER_LIMIT = 200
        # The client will need to split this across more requests in order to
        # stay within the tighter limit.
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = client["wide"].read()
        xarray.testing.assert_equal(actual, expected)
        higher_request_count = len(requests)
        # Tighten even more.
        xarray_client.URL_CHARACTER_LIMIT = 100
        requests.clear()  # Empty the Request cache before the next batch of requests.
        actual = client["wide"].read()
        xarray.testing.assert_equal(actual, expected)
        highest_request_count = len(requests)
    finally:
        # Restore default.
        xarray_client.URL_CHARACTER_LIMIT = original
    # The goal here is to test the *trend* not the specific values because the
    # number of requests may evolve as the library changes, but the trend should
    # hold.
    assert highest_request_count > higher_request_count > normal_request_count



import io

from ..media_type_registration import serialization_registry
from ..utils import modules_available


def walk(node, pre=None):
    """
    Yield (key_path, value) where each value is an ArrayAdapter.

    As a succinct illustration (does not literally run):

    >>> list(walk{"a": {"b": 1, "c": {"d": 2}}})
    [
        (("a", "b"), 1),
        (("a", "c", "d"), 2),
    ]
    """
    pre = pre[:] if pre else []
    if node.structure_family != "array":
        for key, value in node.items():
            for d in walk(value, pre + [key]):
                yield d
    else:
        yield (pre, node)


if modules_available("h5py"):

    def serialize_hdf5(node, metadata):
        """
        Encode everything below this node as HDF5.

        Walk node. Encode all nodes an dataframes as Groups, all arrays and columns as Datasets.
        """
        import h5py

        buffer = io.BytesIO()
        root_node = node
        with h5py.File(buffer, mode="w") as file:
            for k, v in metadata.items():
                file.attrs.create(k, v)
            for key_path, array_adapter in walk(node):
                group = file
                node = root_node
                for key in key_path[:-1]:
                    node = node[key]
                    if key in group:
                        group = group[key]
                    else:
                        group = group.create_group(key)
                        group.attrs.update(node.metadata)
                data = array_adapter.read()
                dataset = group.create_dataset(key_path[-1], data=data)
                for k, v in array_adapter.metadata.items():
                    dataset.attrs.create(k, v)
        return buffer.getbuffer()

    serialization_registry.register("node", "application/x-hdf5", serialize_hdf5)



import enum
import uuid
from datetime import datetime
from typing import Dict, Generic, List, Optional, TypeVar, Union

import pydantic
import pydantic.dataclasses
import pydantic.generics

DataT = TypeVar("DataT")
LinksT = TypeVar("LinksT")
MetaT = TypeVar("MetaT")


class Error(pydantic.BaseModel):
    code: int
    message: str


class Response(pydantic.generics.GenericModel, Generic[DataT, LinksT, MetaT]):
    data: Optional[DataT]
    error: Optional[Error]
    links: Optional[LinksT]
    meta: Optional[MetaT]

    @pydantic.validator("error", always=True)
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v


class PaginationLinks(pydantic.BaseModel):
    self: str
    next: str
    prev: str
    first: str
    last: str


class EntryFields(str, enum.Enum):
    metadata = "metadata"
    structure_family = "structure_family"
    microstructure = "structure.micro"
    macrostructure = "structure.macro"
    count = "count"
    sorting = "sorting"
    specs = "specs"
    none = ""


class StructureFamilies(str, enum.Enum):
    node = "node"
    array = "array"
    dataframe = "dataframe"
    xarray_data_array = "xarray_data_array"
    xarray_dataset = "xarray_dataset"


class Structure(pydantic.BaseModel):
    micro: Optional[dict]
    macro: Optional[dict]


class SortingDirection(int, enum.Enum):
    ASCENDING = 1
    DECENDING = -1


class SortingItem(pydantic.BaseModel):
    key: str
    direction: SortingDirection


class NodeAttributes(pydantic.BaseModel):
    ancestors: List[str]
    structure_family: Optional[StructureFamilies]
    specs: Optional[List[str]]
    metadata: Optional[dict]  # free-form, user-specified dict
    structure: Optional[Structure]
    count: Optional[int]
    sorting: Optional[List[SortingItem]]


AttributesT = TypeVar("AttributesT")
ResourceMetaT = TypeVar("ResourceMetaT")
ResourceLinksT = TypeVar("ResourceLinksT")


class SelfLinkOnly(pydantic.BaseModel):
    self: str


class NodeLinks(pydantic.BaseModel):
    self: str
    search: str
    full: str


class ArrayLinks(pydantic.BaseModel):
    self: str
    full: str
    block: str


class DataFrameLinks(pydantic.BaseModel):
    self: str
    full: str
    partition: str


class XarrayDataArrayLinks(pydantic.BaseModel):
    self: str
    full_variable: str


class XarrayDatasetLinks(pydantic.BaseModel):
    self: str
    full_variable: str
    full_coord: str
    full_dataset: str


resource_links_type_by_structure_family = {
    "node": NodeLinks,
    "array": ArrayLinks,
    "dataframe": DataFrameLinks,
    "xarray_data_array": XarrayDataArrayLinks,
    "xarray_dataset": XarrayDatasetLinks,
}


class EmptyDict(pydantic.BaseModel):
    pass


class NodeMeta(pydantic.BaseModel):
    count: int


class Resource(
    pydantic.generics.GenericModel, Generic[AttributesT, ResourceLinksT, ResourceMetaT]
):
    "A JSON API Resource"
    id: Union[str, uuid.UUID]
    attributes: AttributesT
    links: Optional[ResourceLinksT]
    meta: Optional[ResourceMetaT]


class AccessAndRefreshTokens(pydantic.BaseModel):
    access_token: str
    expires_in: int
    refresh_token: str
    refresh_token_expires_in: int
    token_type: str


class RefreshToken(pydantic.BaseModel):
    refresh_token: str


class AuthenticationMode(str, enum.Enum):
    password = "password"
    external = "external"


class AboutAuthenticationProvider(pydantic.BaseModel):
    provider: str
    mode: AuthenticationMode
    links: Dict[str, str]
    confirmation_message: Optional[str]


class AboutAuthenticationLinks(pydantic.BaseModel):
    whoami: str
    apikey: str
    refresh_session: str
    revoke_session: str
    logout: str


class AboutAuthentication(pydantic.BaseModel):
    required: bool
    providers: List[AboutAuthenticationProvider]
    links: Optional[AboutAuthenticationLinks]


class About(pydantic.BaseModel):
    api_version: int
    library_version: str
    formats: Dict[str, List[str]]
    aliases: Dict[str, Dict[str, List[str]]]
    queries: List[str]
    authentication: AboutAuthentication
    links: Dict[str, str]
    meta: dict


class PrincipalType(str, enum.Enum):
    user = "user"
    service = "service"  # TODO Add support for services.


class Identity(pydantic.BaseModel, orm_mode=True):
    id: pydantic.constr(max_length=255)
    provider: pydantic.constr(max_length=255)
    latest_login: Optional[datetime]


class Role(pydantic.BaseModel, orm_mode=True):
    name: str
    scopes: List[str]
    # principals


class APIKey(pydantic.BaseModel, orm_mode=True):
    first_eight: pydantic.constr(min_length=8, max_length=8)
    expiration_time: Optional[datetime]
    note: Optional[pydantic.constr(max_length=255)]
    scopes: List[str]
    latest_activity: Optional[datetime] = None


class APIKeyWithSecret(APIKey):
    secret: str  # hex-encoded bytes

    @classmethod
    def from_orm(cls, orm, secret):
        return cls(
            first_eight=orm.first_eight,
            expiration_time=orm.expiration_time,
            note=orm.note,
            scopes=orm.scopes,
            latest_activity=orm.latest_activity,
            secret=secret,
        )


class Session(pydantic.BaseModel, orm_mode=True):
    """
    This related to refresh tokens, which have a session uuid ("sid") claim.

    When the client attempts to use a refresh token, we first check
    here to ensure that the "session", which is associated with a chain
    of refresh tokens that came from a single authentication, are still valid.
    """

    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    uuid: uuid.UUID
    expiration_time: datetime
    revoked: bool


class Principal(pydantic.BaseModel, orm_mode=True):
    "Represents a User or Service"
    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    uuid: uuid.UUID
    type: PrincipalType
    identities: List[Identity] = []
    roles: List[Role] = []
    api_keys: List[APIKey] = []
    sessions: List[Session] = []
    latest_activity: Optional[datetime] = None

    @classmethod
    def from_orm(cls, orm, latest_activity=None):
        instance = super().from_orm(orm)
        instance.latest_activity = latest_activity
        return instance


class APIKeyRequestParams(pydantic.BaseModel):
    # Provide an example for expires_in. Otherwise, OpenAPI suggests lifetime=0.
    # If the user is not reading carefully, they will be frustrated when they
    # try to use the instantly-expiring API key!
    expires_in: Optional[int] = pydantic.Field(..., example=600)  # seconds
    scopes: Optional[List[str]] = pydantic.Field(..., example=["inherit"])
    note: Optional[str]



import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import from_config

arr = ArrayAdapter.from_array(numpy.ones((5, 5)))


def tree_a(access_policy):
    return MapAdapter({"A1": arr, "A2": arr}, access_policy=access_policy)


def tree_b(access_policy):
    return MapAdapter({"B1": arr, "B2": arr}, access_policy=access_policy)


@pytest.fixture
def config(tmpdir):
    return {
        "authentication": {
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {"alice": "secret1", "bob": "secret2"}
                    },
                }
            ],
        },
        "database_uri": f"sqlite:///{tmpdir}/tiled.sqlite",
        "access_control": {
            "access_policy": "tiled.adapters.mapping:SimpleAccessPolicy",
            "args": {"access_lists": {"alice": ["a"]}, "provider": "toy"},
        },
        "trees": [
            {
                "tree": f"{__name__}:tree_a",
                "path": "/a",
                "access_control": {
                    "access_policy": "tiled.adapters.mapping:SimpleAccessPolicy",
                    "args": {
                        "provider": "toy",
                        "access_lists": {
                            "alice": ["A2"],
                            # This should have no effect because bob
                            # cannot access the parent node.
                            "bob": ["A1", "A2"],
                        },
                    },
                },
            },
            {"tree": f"{__name__}:tree_b", "path": "/b"},
        ],
    }


def test_top_level_access_control(enter_password, config):
    with enter_password("secret1"):
        alice_client = from_config(config, username="alice", token_cache={})
    with enter_password("secret2"):
        bob_client = from_config(config, username="bob", token_cache={})
    assert "a" in alice_client
    assert "A2" in alice_client["a"]
    assert "A1" not in alice_client["a"]
    assert "b" not in alice_client
    alice_client["a"]["A2"]
    with pytest.raises(KeyError):
        alice_client["b"]
    assert not list(bob_client)
    with pytest.raises(KeyError):
        bob_client["a"]
    with pytest.raises(KeyError):
        bob_client["b"]



import builtins
import hashlib

import numpy
import tifffile

from ..server.object_cache import with_object_cache
from ..structures.array import ArrayMacroStructure, BuiltinDtype


class TiffAdapter:
    """
    Read a TIFF file.

    Examples
    --------

    >>> TiffAdapter("path/to/file.tiff")
    """

    structure_family = "array"

    def __init__(self, path):
        self._file = tifffile.TiffFile(path)
        self._cache_key = (type(self).__module__, type(self).__qualname__, path)

    @property
    def metadata(self):
        # This contains some enums, but Python's built-in JSON serializer
        # handles them fine (converting  to str or int as appropriate).
        return {tag.name: tag.value for tag in self._file.pages[0].tags.values()}

    def read(self, slice=None):
        # TODO Is there support for reading less than the whole array
        # if we only want a slice? I do not think that is possible with a
        # single-page TIFF but I'm not sure. Certainly it *is* possible for
        # multi-page TIFFs.
        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def read_block(self, block, slice=None):
        # For simplicity, this adapter always treat a single TIFF file as one
        # chunk. This could be relaxed in the future.
        if sum(block) != 0:
            raise IndexError(block)

        arr = with_object_cache(self._cache_key, self._file.asarray)
        if slice is not None:
            arr = arr[slice]
        return arr

    def microstructure(self):
        return BuiltinDtype.from_numpy_dtype(self._file.series[0].dtype)

    def macrostructure(self):
        if self._file.is_shaped:
            shape = tuple(self._file.shaped_metadata[0]["shape"])
        else:
            arr = with_object_cache(self._cache_key, self._file.asarray)
            shape = arr.shape
        return ArrayMacroStructure(shape=shape, chunks=tuple((dim,) for dim in shape))


class TiffSequenceAdapter:

    structure_family = "array"

    def __init__(self, seq):
        self._seq = seq
        self._metadata = {}
        self._cache_key = (
            type(self).__module__,
            type(self).__qualname__,
            hashlib.md5(str(seq.files).encode()).hexdigest(),
        )

    @property
    def metadata(self):
        return self._metadata

    def read(self, slice=None):
        """Return a numpy array

        Receives a sequence of values to select from a collection of tiff files that were saved in a folder
        The input order is defined as files --> X slice --> Y slice
        read() can receive one value or one slice to select all the data from one file or a sequence of files;
        or it can receive a tuple of up to three values (int or slice) to select a more specific sequence of pixels
        of a group of images
        """

        # Print("Inside Adapter:", slice)
        if slice is None:
            return with_object_cache(self._cache_key, _safe_asarray, self._seq)
        if isinstance(slice, int):
            # e.g. read(slice=0)
            return with_object_cache(
                self._cache_key + (slice,), _safe_asarray, self._seq, index=slice
            )
        # e.g. read(slice=(...))
        if isinstance(slice, tuple):
            if len(slice) == 0:
                return with_object_cache(self._cache_key, _safe_asarray, self._seq)
            image_axis, *the_rest = slice
            # Could be int or slice
            # (0, slice(...)) or (0,....) are converted to a list
            if isinstance(image_axis, int):
                # e.g. read(slice=(0, ....))
                arr = with_object_cache(
                    self._cache_key + (image_axis,),
                    _safe_asarray,
                    self._seq,
                    index=image_axis,
                )
            if isinstance(image_axis, builtins.slice):
                if image_axis.start is None:
                    slice_start = 0
                else:
                    slice_start = image_axis.start
                if image_axis.step is None:
                    slice_step = 1
                else:
                    slice_step = image_axis.step
                arr = numpy.stack(
                    [
                        with_object_cache(
                            self._cache_key + (i,), _safe_asarray, self._seq, index=i
                        )
                        for i in range(slice_start, image_axis.stop, slice_step)
                    ]
                )
            arr = arr[tuple(the_rest)]
            return arr
        if isinstance(slice, builtins.slice):
            # Check for start and step which can be optional
            if slice.start is None:
                slice_start = 0
            else:
                slice_start = slice.start
            if slice.step is None:
                slice_step = 1
            else:
                slice_step = slice.step
            arr = numpy.stack(
                [
                    with_object_cache(
                        self._cache_key + (i,), _safe_asarray, self._seq, index=i
                    )
                    for i in range(slice_start, slice.stop, slice_step)
                ]
            )
            return arr

    def read_block(self, block, slice=None):
        # Print("Block:", block)
        if block[1:] != (0, 0):
            raise IndexError(block)
        arr = self.read(builtins.slice(block[0], block[0] + 1))
        if slice is not None:
            arr = arr[slice]
        return arr

    def microstructure(self):
        # Assume all files have the same data type
        return BuiltinDtype.from_numpy_dtype(self.read(slice=0).dtype)

    def macrostructure(self):
        shape = (len(self._seq), *self.read(slice=0).shape)
        # print("array shape", shape)
        return ArrayMacroStructure(
            shape=shape,
            # one chunks per underlying TIFF file
            chunks=((1,) * shape[0], (shape[1],), (shape[2],)),
        )


def subdirectory_handler(path):
    """
    Sniff a subdirectory for TIFF sequences.
    """
    # Max fraction of files that do not look like TIFF sequence files
    RELATIVE_THRESHOLD = 0.2
    # Max number of files that do not look like TIFF sequence files
    ABSOLUTE_THRESHOLD = 10
    filepaths = []
    outliers = 0
    for filepath in path.iterdir():
        if not filepath.is_file():
            # Skip directories.
            continue
        if filepath.name.startswith("."):
            # Skip hidden files.
            continue
        if (filepath.suffix in (".tif", ".tiff")) and filepath.stem[-3:].isdigit():
            # This looks like something123.tif.
            filepaths.append(filepath)
            continue
        outliers += 1

    fraction = outliers / (outliers + len(filepaths))
    if (outliers <= ABSOLUTE_THRESHOLD) and (fraction <= RELATIVE_THRESHOLD):
        # This looks like a TIFF sequence directory.
        import tifffile

        seq = tifffile.TiffSequence(sorted(filepaths))
        return TiffSequenceAdapter(seq)

    return None


def _safe_asarray(seq, index=None):
    """
    Wrap FileSequence.asarray to cope with this breaking change in tifffile.

    https://github.com/cgohlke/tifffile/commit/18315846ebf0b31126dfdeffee8e0ef30d81b31c#diff-d8d5cccae6533b861bdf6c0c20c6f49415978effec9a13797909535f3fc7390dL9563-R9746  # noqa
    """

    try:
        return seq.asarray(index=index)
    except TypeError:
        # usage for versions before 2021.10.10
        return seq.asarray(file=index)



import collections.abc
import warnings

import dask.array
import h5py
import numpy

from ..adapters.utils import IndexersMixin, tree_repr
from ..utils import DictView
from .array import ArrayAdapter


class MockHDF5Dataset:
    "Mock just enough of the HDF5 Dataset interface to act as a placeholder."

    def __init__(self, array, attrs):
        self._array = array
        self.shape = array.shape
        self.dtype = array.dtype
        self.ndim = array.ndim
        self.attrs = attrs

    def __getitem__(self, key):
        return self._array.__getitem__(key)


class HDF5DatasetAdapter(ArrayAdapter):
    # TODO Just wrap h5py.Dataset directly, not via dask.array.
    def __init__(self, dataset):
        super().__init__(dask.array.from_array(dataset), metadata=dataset.attrs)


class HDF5Adapter(collections.abc.Mapping, IndexersMixin):
    """
    Read an HDF5 file or a group within one.

    This map the structure of an HDF5 file onto a "Tree" of array structures.

    Examples
    --------

    From the root node of a file given a filepath

    >>> import h5py
    >>> HDF5Adapter.from_file("path/to/file.h5")

    From the root node of a file given an h5py.File object

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter.from_file(file)

    From a group within a file

    >>> import h5py
    >>> file = h5py.File("path/to/file.h5")
    >>> HDF5Adapter(file["some_group']["some_sub_group"])

    """

    structure_family = "node"

    def __init__(self, node, access_policy=None, principal=None):
        if (access_policy is not None) and (
            not access_policy.check_compatibility(self)
        ):
            raise ValueError(
                f"Access policy {access_policy} is not compatible with this Tree."
            )
        self._node = node
        self._access_policy = access_policy
        self._principal = principal
        super().__init__()

    @classmethod
    def from_file(cls, file):
        if not isinstance(file, h5py.File):
            file = h5py.File(file, "r")
        return cls(file)

    def __repr__(self):
        return tree_repr(self, list(self))

    @property
    def access_policy(self):
        return self._access_policy

    @property
    def principal(self):
        return self._principal

    def authenticated_as(self, principal):
        if self._principal is not None:
            raise RuntimeError(f"Already authenticated as {self.principal}")
        if self._access_policy is not None:
            raise NotImplementedError
        tree = type(self)(
            self._node,
            access_policy=self._access_policy,
            principal=principal,
        )
        return tree

    @property
    def metadata(self):
        d = dict(self._node.attrs)
        for k, v in list(d.items()):
            # Convert any bytes to str.
            if isinstance(v, bytes):
                d[k] = v.decode()
        return DictView(d)

    def __iter__(self):
        yield from self._node

    def __getitem__(self, key):
        value = self._node[key]
        if isinstance(value, h5py.Group):
            return HDF5Adapter(value)
        else:
            if value.dtype == numpy.dtype("O"):
                warnings.warn(
                    f"The dataset {key} is of object type, using a "
                    "Python-only feature of h5py that is not supported by "
                    "HDF5 in general. Read more about that feature at "
                    "https://docs.h5py.org/en/stable/special.html. "
                    "Consider using a fixed-length field instead. "
                    "Tiled will serve an empty placeholder, unless the "
                    "object is of size 1, where it will attempt to repackage "
                    "the data into a numpy array."
                )

                check_str_dtype = h5py.check_string_dtype(value.dtype)
                if check_str_dtype.length is None:
                    dataset_names = value.file[self._node.name + "/" + key][...][()]
                    if value.size == 1:
                        arr = MockHDF5Dataset(numpy.array(dataset_names), {})
                        return HDF5DatasetAdapter(arr)
                return HDF5DatasetAdapter(MockHDF5Dataset(numpy.array([]), {}))
            return HDF5DatasetAdapter(value)

    def __len__(self):
        return len(self._node)

    def search(self, query):
        """
        Return a Tree with a subset of the mapping.
        """
        raise NotImplementedError

    def read(self, fields=None):
        if fields is not None:
            raise NotImplementedError
        return self

    # The following three methods are used by IndexersMixin
    # to define keys_indexer, items_indexer, and values_indexer.

    def _keys_slice(self, start, stop, direction):
        keys = list(self._node)
        if direction < 0:
            keys = reversed(keys)
        return keys[start:stop]

    def _items_slice(self, start, stop, direction):
        items = [(key, self[key]) for key in list(self)]
        if direction < 0:
            items = reversed(items)
        return items[start:stop]

    def _item_by_index(self, index, direction):
        keys = list(self)
        if direction < 0:
            keys = reversed(keys)
        return keys[index]



import importlib

from ..utils import UNCHANGED, DictView, ListView, OneShotCachedMap
from .cache import Revalidate, verify_cache


class BaseClient:
    def __init__(self, context, *, item, path, params, structure_clients):
        self._context = context
        if isinstance(path, str):
            raise ValueError("path is expected to be a list of segments")
        # Stash *immutable* copies just to be safe.
        self._path = tuple(path or [])
        self._item = item
        self._cached_len = None  # a cache just for __len__
        self._params = params or {}
        self.structure_clients = structure_clients
        super().__init__()

    def login(self, provider=None):
        """
        Depending on the server's authentication method, this will prompt for username/password

        >>> c.login()
        Username: USERNAME
        Password: <input is hidden>

        or prompt you to open a link in a web browser to login with a third party and paste in a access code

        >>> c.login()
        Navigate your web browser to this address to obtain access code:

        ...

        Access code (quotes optional): <input is hidden>

        See also c.context.authenticate() and c.context.reauthenticate().
        """
        self.context.authenticate(provider=provider)
        # Do NOT return the tokens that are returned by authenticate().
        # This avoids displaying valid refresh tokens into places they might persist,
        # like Jupyter notebooks.

    def logout(self):
        """
        Log out.
        """
        self.context.logout()

    def __repr__(self):
        return f"<{type(self).__name__}>"

    @property
    def context(self):
        return self._context

    @property
    def item(self):
        "JSON payload describing this item. Mostly for internal use."
        return self._item

    @property
    def metadata(self):
        "Metadata about this data source."
        # Ensure this is immutable (at the top level) to help the user avoid
        # getting the wrong impression that editing this would update anything
        # persistent.
        return DictView(self._item["attributes"]["metadata"])

    @property
    def path(self):
        "Sequence of entry names from the root Tree to this entry"
        return ListView(self._path)

    @property
    def uri(self):
        "Direct link to this entry"
        return self.item["links"]["self"]

    @property
    def username(self):
        return self.context.username

    @property
    def offline(self):
        return self.context.offline

    @offline.setter
    def offline(self, value):
        self.context.offline = bool(value)

    def new_variation(
        self, path=UNCHANGED, params=UNCHANGED, structure_clients=UNCHANGED, **kwargs
    ):
        """
        This is intended primarily for internal use and use by subclasses.
        """
        if path is UNCHANGED:
            path = self._path
        if params is UNCHANGED:
            params = self._params
        if structure_clients is UNCHANGED:
            structure_clients = self.structure_clients
        return type(self)(
            item=self._item,
            path=path,
            params=params,
            structure_clients=structure_clients,
            **kwargs,
        )


class BaseStructureClient(BaseClient):
    def __init__(self, *args, structure=None, **kwargs):
        super().__init__(*args, **kwargs)
        if structure is not None:
            # Allow the caller to optionally hand us a structure that is already
            # parsed from a dict into a structure dataclass.
            self._structure = structure
        else:
            attributes = self.item["attributes"]
            structure_type = STRUCTURE_TYPES[attributes["structure_family"]]
            self._structure = structure_type.from_json(attributes["structure"])

    def download(self):
        """
        Download all data into the cache.

        This causes it to be cached if the context is configured with a cache.
        """
        verify_cache(self.context.cache)
        repr(self)
        self.read()

    def refresh(self, force=False):
        """
        Refresh cached data for this node.

        Parameters
        ----------
        force: bool
            If False, (default) refresh only expired cache entries.
            If True, refresh all cache entries.
        """
        if force:
            revalidate = Revalidate.FORCE
        else:
            revalidate = Revalidate.IF_EXPIRED
        with self.context.revalidation(revalidate):
            self.download()

    def structure(self):
        """
        Return a dataclass describing the structure of the data.
        """
        if self._structure.macro.resizable:
            # In the future, conditionally fetch updated information.
            raise NotImplementedError(
                "The server has indicated that this has a dynamic, resizable "
                "structure and this version of the Tiled Python client cannot "
                "cope with that."
            )
        return self._structure


STRUCTURE_TYPES = OneShotCachedMap(
    {
        "array": lambda: importlib.import_module(
            "...structures.array", BaseStructureClient.__module__
        ).ArrayStructure,
        "dataframe": lambda: importlib.import_module(
            "...structures.dataframe", BaseStructureClient.__module__
        ).DataFrameStructure,
        "xarray_data_array": lambda: importlib.import_module(
            "...structures.xarray", BaseStructureClient.__module__
        ).DataArrayStructure,
        "xarray_dataset": lambda: importlib.import_module(
            "...structures.xarray", BaseStructureClient.__module__
        ).DatasetStructure,
    }
)



import logging
import secrets

from fastapi import APIRouter, Request
from jose import JWTError, jwk, jwt
from starlette.responses import RedirectResponse

from .server.authentication import Mode
from .utils import modules_available

logger = logging.getLogger(__name__)


class DummyAuthenticator:
    """
    For test and demo purposes only!

    Accept any username and any password.

    """

    mode = Mode.password

    async def authenticate(self, username: str, password: str):
        return username


class DictionaryAuthenticator:
    """
    For test and demo purposes only!

    Check passwords from a dictionary of usernames mapped to passwords.
    """

    mode = Mode.password
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  users_to_password:
    type: object
  description: |
    Mapping usernames to password. Environment variable expansion should be
    used to avoid placing passwords directly in configuration.
"""

    def __init__(self, users_to_passwords):
        self._users_to_passwords = users_to_passwords

    async def authenticate(self, username: str, password: str):
        true_password = self._users_to_passwords.get(username)
        if not true_password:
            # Username is not valid.
            return
        if secrets.compare_digest(true_password, password):
            return username


class PAMAuthenticator:

    mode = Mode.password
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  service:
    type: string
    description: PAM service. Default is 'login'.
"""

    def __init__(self, service="login"):
        if not modules_available("pamela"):
            raise ModuleNotFoundError(
                "This PAMAuthenticator requires the module 'pamela' to be installed."
            )
        self.service = service
        # TODO Try to open a PAM session.

    async def authenticate(self, username: str, password: str):
        import pamela

        try:
            pamela.authenticate(username, password, service=self.service)
        except pamela.PAMError:
            # Authentication failed.
            return
        else:
            return username


class OIDCAuthenticator:

    mode = Mode.external
    configuration_schema = """
$schema": http://json-schema.org/draft-07/schema#
type: object
additionalProperties: false
properties:
  client_id:
    type: string
  client_secret:
    type: string
  redirect_uri:
    type: string
  token_uri:
    type: string
  authorization_endpoint:
    type: string
  public_keys:
    type: array
    item:
      type: object
      properties:
        - alg:
            type: string
        - e
            type: string
        - kid
            type: string
        - kty
            type: string
        - n
            type: string
        - use
            type: string
      required:
        - alg
        - e
        - kid
        - kty
        - n
        - use
"""

    def __init__(
        self,
        client_id,
        client_secret,
        redirect_uri,
        public_keys,
        token_uri,
        authorization_endpoint,
        confirmation_message,
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.confirmation_message = confirmation_message
        self.redirect_uri = redirect_uri
        self.public_keys = public_keys
        self.token_uri = token_uri
        self.authorization_endpoint = authorization_endpoint.format(
            client_id=client_id, redirect_uri=redirect_uri
        )

    async def authenticate(self, request):
        code = request.query_params["code"]
        response = await exchange_code(
            self.token_uri, code, self.client_id, self.client_secret, self.redirect_uri
        )
        response_body = response.json()
        if response.is_error:
            logger.error("Authentication error: %r", response_body)
            return None
        response_body = response.json()
        id_token = response_body["id_token"]
        access_token = response_body["access_token"]
        # Match the kid in id_token to a key in the list of public_keys.
        key = find_key(id_token, self.public_keys)
        try:
            verified_body = jwt.decode(
                id_token, key, access_token=access_token, audience=self.client_id
            )
        except JWTError:
            logger.exception(
                "Authentication error. Unverified token: %r",
                jwt.get_unverified_claims(id_token),
            )
            return None
        return verified_body["sub"]


class KeyNotFoundError(Exception):
    pass


def find_key(token, keys):
    """
    Find a key from the configured keys based on the kid claim of the token

    Parameters
    ----------
    token : token to search for the kid from
    keys:  list of keys

    Raises
    ------
    KeyNotFoundError:
        returned if the token does not have a kid claim

    Returns
    ------
    key: found key object
    """

    unverified = jwt.get_unverified_header(token)
    kid = unverified.get("kid")
    if not kid:
        raise KeyNotFoundError("No 'kid' in token")

    for key in keys:
        if key["kid"] == kid:
            return jwk.construct(key)
    return KeyNotFoundError(
        f"Token specifies {kid} but we have {[k['kid'] for k in keys]}"
    )


async def exchange_code(token_uri, auth_code, client_id, client_secret, redirect_uri):
    """Method that talks to an IdP to exchange a code for an access_token and/or id_token
    Args:
        token_url ([type]): [description]
        auth_code ([type]): [description]
    """
    if not modules_available("httpx"):
        raise ModuleNotFoundError(
            "This authenticator requires 'httpx'. (pip install httpx)"
        )
    import httpx

    response = httpx.post(
        url=token_uri,
        data={
            "grant_type": "authorization_code",
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "code": auth_code,
            "client_secret": client_secret,
        },
    )
    return response


class SAMLAuthenticator:

    mode = Mode.external

    def __init__(
        self,
        saml_settings,  # See EXAMPLE_SAML_SETTINGS below.
        attribute_name,  # which SAML attribute to use as 'id' for Idenity
        confirmation_message=None,
    ):
        self.saml_settings = saml_settings
        self.attribute_name = attribute_name
        self.confirmation_message = confirmation_message
        self.authorization_endpoint = "/login"

        router = APIRouter()

        if not modules_available("onelogin"):
            # The PyPI package name is 'python3-saml'
            # but it imports as 'onelogin'.
            # https://github.com/onelogin/python3-saml
            raise ModuleNotFoundError(
                "This SAMLAuthenticator requires 'python3-saml' to be installed."
            )

        from onelogin.saml2.auth import OneLogin_Saml2_Auth

        @router.get("/login")
        async def saml_login(request: Request):
            req = await prepare_saml_from_fastapi_request(request)
            auth = OneLogin_Saml2_Auth(req, self.saml_settings)
            # saml_settings = auth.get_settings()
            # metadata = saml_settings.get_sp_metadata()
            # errors = saml_settings.validate_metadata(metadata)
            # if len(errors) == 0:
            #   print(metadata)
            # else:
            #   print("Error found on Metadata: %s" % (', '.join(errors)))
            callback_url = auth.login()
            response = RedirectResponse(url=callback_url)
            return response

        self.include_routers = [router]

    async def authenticate(self, request):
        if not modules_available("onelogin"):
            raise ModuleNotFoundError(
                "This SAMLAuthenticator requires the module 'oneline' to be installed."
            )
        from onelogin.saml2.auth import OneLogin_Saml2_Auth

        req = await prepare_saml_from_fastapi_request(request, True)
        auth = OneLogin_Saml2_Auth(req, self.saml_settings)
        auth.process_response()  # Process IdP response
        errors = auth.get_errors()  # This method receives an array with the errors
        if errors:
            raise Exception(
                "Error when processing SAML Response: %s %s"
                % (", ".join(errors), auth.get_last_error_reason())
            )
        if auth.is_authenticated():
            # Return a string that the Identity can use as id.
            attribute_as_list = auth.get_attributes()[self.attribute_name]
            # Confused in what situation this would have more than one item....
            assert len(attribute_as_list) == 1
            return attribute_as_list[0]
        else:
            return None


async def prepare_saml_from_fastapi_request(request, debug=False):
    form_data = await request.form()
    rv = {
        "http_host": request.client.host,
        "server_port": request.url.port,
        "script_name": request.url.path,
        "post_data": {},
        "get_data": {}
        # Advanced request options
        # "https": "",
        # "request_uri": "",
        # "query_string": "",
        # "validate_signature_from_qs": False,
        # "lowercase_urlencoding": False
    }
    if request.query_params:
        rv["get_data"] = (request.query_params,)
    if "SAMLResponse" in form_data:
        SAMLResponse = form_data["SAMLResponse"]
        rv["post_data"]["SAMLResponse"] = SAMLResponse
    if "RelayState" in form_data:
        RelayState = form_data["RelayState"]
        rv["post_data"]["RelayState"] = RelayState
    return rv



import io
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import xarray

from ..media_type_registration import deserialization_registry, serialization_registry
from ..utils import modules_available
from .array import ArrayStructure
from .dataframe import (
    APACHE_ARROW_FILE_MIME_TYPE,
    XLSX_MIME_TYPE,
    serialize_arrow,
    serialize_csv,
    serialize_excel,
    serialize_html,
    serialize_parquet,
)


@dataclass
class DataArrayMacroStructure:
    variable: ArrayStructure
    coords: Optional[
        Dict[str, str]
    ]  # overridden below to be Optional[Dict[str, DataArrayStructure]]
    coord_names: List[str]
    name: str
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_json(cls, structure):
        if structure["coords"] is not None:
            coords = {
                key: DataArrayStructure.from_json(value)
                for key, value in structure["coords"].items()
            }
        else:
            coords = None
        return cls(
            variable=ArrayStructure.from_json(structure["variable"]),
            coords=coords,
            coord_names=structure["coord_names"],
            name=structure["name"],
        )


@dataclass
class DataArrayStructure:
    macro: DataArrayMacroStructure
    micro: None

    @classmethod
    def from_json(cls, structure):
        return cls(
            macro=DataArrayMacroStructure.from_json(structure["macro"]), micro=None
        )


# Define a nested structure now that the necessary object has been defined.
DataArrayMacroStructure.__annotations__[
    "coords"
] = DataArrayMacroStructure.__annotations__["coords"].copy_with(
    (str, DataArrayMacroStructure)
)


@dataclass
class DatasetMacroStructure:
    data_vars: Dict[str, DataArrayStructure]
    coords: Dict[str, DataArrayStructure]
    resizable: Union[bool, Tuple[bool, ...]] = False

    @classmethod
    def from_json(cls, structure):
        return cls(
            data_vars={
                key: DataArrayStructure.from_json(value)
                for key, value in structure["data_vars"].items()
            },
            coords={
                key: DataArrayStructure.from_json(value)
                for key, value in structure["coords"].items()
            },
        )


@dataclass
class DatasetStructure:
    macro: DatasetMacroStructure
    micro: None

    @classmethod
    def from_json(cls, structure):
        return cls(
            macro=DatasetMacroStructure.from_json(structure["macro"]), micro=None
        )


class _BytesIOThatIgnoresClose(io.BytesIO):
    def close(self):
        # When the netcdf writer tells us to close(), ignore it.
        pass


if modules_available("h5netcdf"):

    def serialize_netcdf(dataset, metadata):
        file = _BytesIOThatIgnoresClose()
        # This engine is reportedly faster.
        # Also, by avoiding the default engine, we avoid a dependency on 'scipy'.
        dataset.to_netcdf(file, engine="h5netcdf")
        return file.getbuffer()

    # Both application/netcdf and application/x-netcdf are used.
    # https://en.wikipedia.org/wiki/NetCDF
    serialization_registry.register(
        "xarray_dataset", "application/netcdf", serialize_netcdf
    )
    serialization_registry.register(
        "xarray_dataset", "application/x-netcdf", serialize_netcdf
    )

# Support DataFrame formats by first converting to DataFrame.
# This doesn't make much sense for N-dimensional variables, but for
# 1-dimensional variables it is useful.
serialization_registry.register(
    "xarray_dataset",
    APACHE_ARROW_FILE_MIME_TYPE,
    lambda ds, metadata: serialize_arrow(ds.to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "application/x-parquet",
    lambda ds, metadata: serialize_parquet(ds.to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/csv",
    lambda ds, metadata: serialize_csv(ds.to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/plain",
    lambda ds, metadata: serialize_csv(ds.to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    "text/html",
    lambda ds, metadata: serialize_html(ds.to_dataframe(), metadata),
)
serialization_registry.register(
    "xarray_dataset",
    XLSX_MIME_TYPE,
    lambda ds, metadata: serialize_excel(ds.to_dataframe(), metadata),
)
if modules_available("orjson"):
    import orjson

    def serialize_json(ds, metadata):
        df = ds.to_dataframe()
        return orjson.dumps(
            {column: df[column].tolist() for column in df},
        )

    serialization_registry.register(
        "xarray_dataset",
        "application/json",
        serialize_json,
    )

deserialization_registry.register(
    "xarray_dataset", "application/x-zarr", lambda ds, metadata: xarray.open_zarr(ds)
)
# TODO How should we add support for access via Zarr?



import contextlib
import os.path
import tempfile

_here = os.path.abspath(os.path.dirname(__file__))

ALEMBIC_INI_TEMPLATE_PATH = os.path.join(_here, "alembic.ini.template")
ALEMBIC_DIR = os.path.join(_here, "migrations")


def write_alembic_ini(path, database_uri):
    """Write a complete alembic.ini from our template.

    Parameters
    ----------
    path : str
        path to the alembic.ini file that should be written.
    db_url : str
        The SQLAlchemy database url, e.g. `sqlite:///tiled.sqlite`.
    """
    with open(ALEMBIC_INI_TEMPLATE_PATH) as f:
        alembic_ini_tpl = f.read()

    with open(path, "w") as f:
        f.write(
            alembic_ini_tpl.format(
                migration_script_directory=ALEMBIC_DIR,
                # If there are any %s in the URL, they should be replaced with %%, since ConfigParser
                # by default uses %() for substitution. You'll get %s in your URL when you have usernames
                # with special chars (such as '@') that need to be URL encoded. URL Encoding is done with %s.
                # YAY for nested templates?
                database_uri=str(database_uri).replace("%", "%%"),
            )
        )


@contextlib.contextmanager
def temp_alembic_ini(database_uri):
    """
    Context manager for temporary alembic configuration file

    Temporarily write an alembic.ini file for use with alembic migration scripts.
    Context manager yields alembic.ini path.

    Parameters
    ----------
    datbase_uri : str
        The SQLAlchemy database url, e.g. `sqlite:///tiled.sqlite`.

    Returns
    -------
    alembic_ini: str
        The path to the temporary alembic.ini that we have created.
        This file will be cleaned up on exit from the context manager.
    """
    with tempfile.TemporaryDirectory() as td:
        alembic_ini = os.path.join(td, "alembic.ini")
        write_alembic_ini(alembic_ini, database_uri)
        yield alembic_ini



import contextlib
import io
import time
from datetime import timedelta

import numpy
import pytest

from ..adapters.array import ArrayAdapter
from ..adapters.mapping import MapAdapter
from ..client import from_config
from ..client.context import CannotRefreshAuthentication
from ..client.utils import ClientError

arr = ArrayAdapter.from_array(numpy.ones((5, 5)))


tree = MapAdapter({"A1": arr, "A2": arr})


@contextlib.contextmanager
def fail_with_status_code(status_code):
    with pytest.raises(ClientError) as info:
        yield
    assert info.value.response.status_code == status_code


@pytest.fixture
def config(tmpdir):
    """
    Return config with

    - a unique temporary sqlite database location
    - a unique nested dict instance that the test can mutate
    """
    return {
        "authentication": {
            "secret_keys": ["SECRET"],
            "providers": [
                {
                    "provider": "toy",
                    "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                    "args": {
                        "users_to_passwords": {"alice": "secret1", "bob": "secret2"}
                    },
                }
            ],
        },
        "database_uri": f"sqlite:///{tmpdir}/tiled.sqlite",
        "trees": [
            {
                "tree": f"{__name__}:tree",
                "path": "/",
            },
        ],
    }


def test_password_auth(enter_password, config):
    """
    A password that is wrong, empty, or belonging to a different user fails.
    """

    with enter_password("secret1"):
        from_config(config, username="alice", token_cache={})
    with enter_password("secret2"):
        from_config(config, username="bob", token_cache={})

    # Bob's password should not work for alice
    with fail_with_status_code(401):
        with enter_password("secret2"):
            from_config(config, username="alice", token_cache={})

    # Empty password should not work.
    with fail_with_status_code(422):
        with enter_password(""):
            from_config(config, username="alice", token_cache={})


def test_key_rotation(enter_password, config):
    """
    Rotate in a new secret used to sign keys.
    Confirm that clients experience a smooth transition.
    """

    # Obtain refresh token.
    token_cache = {}
    with enter_password("secret1"):
        from_config(config, username="alice", token_cache=token_cache)
    # Use refresh token (no prompt to reauthenticate).
    from_config(config, username="alice", token_cache=token_cache)

    # Rotate in a new key.
    config["authentication"]["secret_keys"].insert(0, "NEW_SECRET")
    assert config["authentication"]["secret_keys"] == ["NEW_SECRET", "SECRET"]
    # The refresh token from the old key is still valid.
    # We reauthenticate and receive a refresh token for the new key.
    from_config(config, username="alice", token_cache=token_cache)

    # Rotate out the old key.
    del config["authentication"]["secret_keys"][1]
    assert config["authentication"]["secret_keys"] == ["NEW_SECRET"]
    # New refresh token works with the new key.
    from_config(config, username="alice", token_cache=token_cache)


def test_refresh_flow(enter_password, config):
    """
    Run a server with an artificially short max access token age
    to force a refresh.
    """

    # Normal default configuration: a refresh is not immediately required.
    token_cache = {}
    with enter_password("secret1"):
        client = from_config(config, username="alice", token_cache=token_cache)
    token1 = client.context.tokens["access_token"]
    client["A1"]
    assert token1 is client.context.tokens["access_token"]

    # Forcing a refresh gives us a new token.
    client.context.reauthenticate()
    token2 = client.context.tokens["access_token"]
    assert token2 is not token1

    # Pathological configuration: a refresh is almost immediately required
    config["authentication"]["access_token_max_age"] = timedelta(seconds=1)
    token_cache = {}
    with enter_password("secret1"):
        client = from_config(config, username="alice", token_cache=token_cache)
    token3 = client.context.tokens["access_token"]
    time.sleep(2)
    # A refresh should happen automatically now.
    client["A1"]
    token4 = client.context.tokens["access_token"]
    assert token3 is not token4

    # Pathological configuration: sessions do not last
    config["authentication"]["session_max_age"] = timedelta(seconds=1)
    token_cache = {}
    with enter_password("secret1"):
        client = from_config(config, username="alice", token_cache=token_cache)
    time.sleep(2)
    # Refresh should fail because the session is too old.
    with pytest.raises(CannotRefreshAuthentication):
        # Set prompt=False so that this raises instead of interactively prompting.
        client.context.reauthenticate(prompt=False)


def test_revoke_session(enter_password, config):
    with enter_password("secret1"):
        client = from_config(config, username="alice", token_cache={})
    # Get the current session ID.
    info = client.context.whoami()
    (session,) = info["sessions"]
    assert not session["revoked"]
    # Revoke it.
    client.context.revoke_session(session["uuid"])
    # Update info and confirm it is listed as revoked.
    updated_info = client.context.whoami()
    (updated_session,) = updated_info["sessions"]
    assert updated_session["revoked"]
    # Confirm it cannot be refreshed.
    with pytest.raises(CannotRefreshAuthentication):
        # Set prompt=False so that this raises instead of interactively prompting.
        client.context.reauthenticate(prompt=False)


def test_multiple_providers(enter_password, config, monkeypatch):
    """
    Test a configuration with multiple identity providers.

    This mechanism is used to support "Login with ORCID or Google or ...."
    """
    config["authentication"]["providers"].extend(
        [
            {
                "provider": "second",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {"users_to_passwords": {"cara": "secret3", "doug": "secret4"}},
            },
            {
                "provider": "third",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {
                    # Duplicate 'cara' username.
                    "users_to_passwords": {"cara": "secret5", "emilia": "secret6"}
                },
            },
        ],
    )
    monkeypatch.setattr("sys.stdin", io.StringIO("1\n"))
    with enter_password("secret1"):
        client = from_config(config, username="alice", token_cache={})
    client.context.whoami()
    monkeypatch.setattr("sys.stdin", io.StringIO("2\n"))
    with enter_password("secret3"):
        client = from_config(config, username="cara", token_cache={})
    monkeypatch.setattr("sys.stdin", io.StringIO("3\n"))
    with enter_password("secret5"):
        client = from_config(config, username="cara", token_cache={})
    client.context.whoami()


def test_multiple_providers_name_collision(config):
    """
    Check that we enforce unique provider names.
    """
    config["authentication"]["providers"] = [
        {
            "provider": "some_name",
            "authenticator": "tiled.authenticators:DictionaryAuthenticator",
            "args": {"users_to_passwords": {"cara": "secret3", "doug": "secret4"}},
        },
        {
            "provider": "some_name",  # duplicate!
            "authenticator": "tiled.authenticators:DictionaryAuthenticator",
            "args": {
                # Duplicate 'cara' username.
                "users_to_passwords": {"cara": "secret5", "emilia": "secret6"}
            },
        },
    ]
    with pytest.raises(ValueError):
        from_config(config)


def test_admin(enter_password, config):
    """
    Test that the 'tiled_admin' config confers the 'admin' Role on a Principal.
    """
    # Make alice an admin. Leave bob as a user.
    config["authentication"]["tiled_admins"] = [{"provider": "toy", "id": "alice"}]

    with enter_password("secret1"):
        admin_client = from_config(config, username="alice", token_cache={})

    with enter_password("secret2"):
        user_client = from_config(config, username="bob", token_cache={})

    user_roles = user_client.context.whoami()["roles"]
    assert [role["name"] for role in user_roles] == ["user"]

    adming_roles = admin_client.context.whoami()["roles"]
    assert "admin" in [role["name"] for role in adming_roles]


def test_api_keys(enter_password, config):
    """
    Test creating, revoking, expiring API keys.
    Test that they have appropriate scope-limited access.
    """
    # Make alice an admin. Leave bob as a user.
    config["authentication"]["tiled_admins"] = [{"provider": "toy", "id": "alice"}]

    with enter_password("secret1"):
        admin_client = from_config(config, username="alice", token_cache={})

    with enter_password("secret2"):
        user_client = from_config(config, username="bob", token_cache={})

    # Make and use an API key. Check that latest_activity is updated.
    user_key_info = user_client.context.create_api_key()
    assert user_key_info["latest_activity"] is None  # never used
    user_client_from_key = from_config(config, api_key=user_key_info["secret"])
    # Check that api_key property is set.
    assert user_client_from_key.context.api_key == user_key_info["secret"]
    # Use the key for a couple requests and see that latest_activity becomes set and then increases.
    user_client_from_key["A1"]
    key_activity1 = user_client_from_key.context.which_api_key()["latest_activity"]
    principal_activity1 = user_client_from_key.context.whoami()["latest_activity"]
    assert key_activity1 is not None
    time.sleep(2)  # Ensure time resolution (1 second) has ticked up.
    user_client_from_key["A1"]
    key_activity2 = user_client_from_key.context.which_api_key()["latest_activity"]
    principal_activity2 = user_client_from_key.context.whoami()["latest_activity"]
    assert key_activity2 > key_activity1
    assert principal_activity2 > principal_activity1
    assert len(user_client_from_key.context.whoami()["api_keys"]) == 1

    # Unset the API key.
    secret = user_client_from_key.context.api_key
    user_client_from_key.context.api_key = None
    with pytest.raises(RuntimeError):
        user_client_from_key.context.which_api_key()
    # Set the API key.
    user_client_from_key.context.api_key = secret
    # Now this works again.
    user_client_from_key.context.which_api_key()

    # Request a key with reduced scope that cannot read metadata.
    admin_key_info = admin_client.context.create_api_key(scopes=["metrics"])
    with fail_with_status_code(401):
        from_config(config, api_key=admin_key_info["secret"])

    # Request a key with reduced scope that can *only* read metadata.
    admin_key_info = admin_client.context.create_api_key(scopes=["read:metadata"])
    restricted_client = from_config(config, api_key=admin_key_info["secret"])
    restricted_client["A1"]
    with fail_with_status_code(401):
        restricted_client["A1"].read()  # no 'read:data' scope

    # Try to request a key with more scopes that the user has.
    with fail_with_status_code(400):
        user_client.context.create_api_key(scopes=["admin:apikeys"])

    # Create and revoke key.
    user_key_info = user_client.context.create_api_key(note="will revoke soon")
    assert len(user_client_from_key.context.whoami()["api_keys"]) == 2
    # There should now be two keys, one from above and this new one, with our note.
    for api_key in user_client_from_key.context.whoami()["api_keys"]:
        if api_key["note"] == "will revoke soon":
            break
    else:
        assert False, "No api keys had a matching note."
    # Revoke the new key.
    user_client_from_key.context.revoke_api_key(user_key_info["first_eight"])
    with fail_with_status_code(401):
        from_config(config, api_key=user_key_info["secret"])
    assert len(user_client_from_key.context.whoami()["api_keys"]) == 1

    # Create a key with a very short lifetime.
    user_key_info = user_client.context.create_api_key(
        note="will expire very soon", expires_in=1
    )  # units: seconds
    time.sleep(2)
    with fail_with_status_code(401):
        from_config(config, api_key=user_key_info["secret"])



import base64
import collections.abc
import dataclasses
import itertools
import math
import operator
import re
import sys
import uuid
from collections import defaultdict
from datetime import datetime, timedelta
from hashlib import md5
from typing import Any

import dateutil.tz
import jmespath
import msgpack
from fastapi import HTTPException, Response
from starlette.responses import JSONResponse, Send, StreamingResponse

# Some are not directly used, but they register things on import.
from .. import queries
from ..adapters.mapping import MapAdapter
from ..queries import KeyLookup, QueryValueError
from ..structures import node  # noqa: F401
from ..structures.dataframe import serialize_arrow
from ..utils import (
    APACHE_ARROW_FILE_MIME_TYPE,
    SerializationError,
    UnsupportedShape,
    modules_available,
    safe_json_dump,
)
from . import schemas
from .etag import tokenize
from .utils import record_timing

del queries
if modules_available("numpy", "dask.array"):
    from ..structures import array as _array  # noqa: F401

    del _array
if modules_available("pandas", "pyarrow", "dask.dataframe"):
    from ..structures import dataframe as _dataframe  # noqa: F401

    del _dataframe
if modules_available("xarray"):
    from ..structures import xarray as _xarray  # noqa: F401

    del _xarray


_FILTER_PARAM_PATTERN = re.compile(r"filter___(?P<name>.*)___(?P<field>[^\d\W][\w\d]+)")
_LOCAL_TZINFO = dateutil.tz.gettz()


def len_or_approx(tree):
    """
    Prefer approximate length if implemented. (It's cheaper.)
    """
    try:
        return operator.length_hint(tree)
    except TypeError:
        return len(tree)


def pagination_links(route, path_parts, offset, limit, length_hint):
    path_str = "/".join(path_parts)
    links = {
        "self": f"{route}/{path_str}?page[offset]={offset}&page[limit]={limit}",
        # These are conditionally overwritten below.
        "first": None,
        "last": None,
        "next": None,
        "prev": None,
    }
    if limit:
        last_page = math.floor(length_hint / limit) * limit
        links.update(
            {
                "first": f"{route}/{path_str}?page[offset]={0}&page[limit]={limit}",
                "last": f"{route}/{path_str}?page[offset]={last_page}&page[limit]={limit}",
            }
        )
    if offset + limit < length_hint:
        links[
            "next"
        ] = f"{route}/{path_str}?page[offset]={offset + limit}&page[limit]={limit}"
    if offset > 0:
        links[
            "prev"
        ] = f"{route}/{path_str}?page[offset]={max(0, offset - limit)}&page[limit]={limit}"
    return links


def construct_entries_response(
    query_registry,
    tree,
    route,
    path,
    offset,
    limit,
    fields,
    select_metadata,
    omit_links,
    filters,
    sort,
    base_url,
    media_type,
):
    path_parts = [segment for segment in path.split("/") if segment]
    if tree.structure_family != "node":
        raise WrongTypeForRoute("This is not a Node; it does not have entries.")
    queries = defaultdict(
        dict
    )  # e.g. {"text": {"text": "dog"}, "lookup": {"key": "..."}}
    # Group the parameters by query type.
    for key, value in filters.items():
        if value is None:
            continue
        name, field = _FILTER_PARAM_PATTERN.match(key).groups()
        queries[name][field] = value
    sorting = []
    if sort is not None:
        for item in sort.split(","):
            if item:
                if item.startswith("-"):
                    sorting.append((item[1:], -1))
                else:
                    sorting.append((item, 1))
    if sorting:
        if not hasattr(tree, "sort"):
            raise HTTPException(
                status_code=400, detail="This Tree does not support sorting."
            )
        tree = tree.sort(sorting)
    # Apply the queries and obtain a narrowed tree.
    key_lookups = []
    for query_name, parameters_dict_of_lists in queries.items():
        for i in itertools.count(0):
            try:
                parameters = {
                    field_name: parameters_list[i]
                    for field_name, parameters_list in parameters_dict_of_lists.items()
                }
            except IndexError:
                break
            query_class = query_registry.name_to_query_type[query_name]
            # Special case:
            # List fields are serialized as comma-separated strings.
            for field in dataclasses.fields(query_class):
                if getattr(field.type, "__origin__", None) is list:
                    (inner_type,) = field.type.__args__
                    parameters[field.name] = [
                        inner_type(item) for item in parameters[field.name].split(",")
                    ]
            try:
                query = query_class(**parameters)
                # Special case: Do key-lookups at the end after all other filtering.
                # We do not require trees to implement this query; we implement it
                # directly here by just calling __getitem__.
                if isinstance(query, KeyLookup):
                    key_lookups.append(query.key)
                    continue
                tree = tree.search(query)
            except QueryValueError as err:
                raise HTTPException(status_code=400, detail=err.args[0])
    if key_lookups:
        # Duplicates are technically legal because *any* query can be given
        # with multiple parameters.
        unique_key_lookups = set(key_lookups)
        (key_lookup), *others = unique_key_lookups
        if others:
            # Two non-equal KeyLookup queries must return no results.
            tree = MapAdapter({})
        else:
            try:
                tree = MapAdapter({key_lookup: tree[key_lookup]}, must_revalidate=False)
            except KeyError:
                tree = MapAdapter({})
    count = len_or_approx(tree)
    links = pagination_links(route, path_parts, offset, limit, count)
    data = []
    if fields != [schemas.EntryFields.none]:
        # Pull a page of items into memory.
        items = tree.items_indexer[offset : offset + limit]  # noqa: E203
    else:
        # Pull a page of just the keys, which is cheaper.
        items = (
            (key, None)
            for key in tree.keys_indexer[offset : offset + limit]  # noqa: E203
        )
    # This value will not leak out. It just used to seed comparisons.
    metadata_stale_at = datetime.utcnow() + timedelta(days=1_000_000)
    must_revalidate = getattr(tree, "must_revalidate", True)
    for key, entry in items:
        resource = construct_resource(
            base_url,
            path_parts + [key],
            entry,
            fields,
            select_metadata,
            omit_links,
            media_type,
        )
        data.append(resource)
        # If any entry has emtry.metadata_stale_at = None, then there will
        # be no 'Expires' header. We will pessimistically assume the values
        # are immediately stale.
        if metadata_stale_at is not None:
            if getattr(entry, "metadata_stale_at", None) is None:
                metadata_stale_at = None
            else:
                metadata_stale_at = min(metadata_stale_at, entry.metadata_stale_at)
    return (
        schemas.Response(data=data, links=links, meta={"count": count}),
        metadata_stale_at,
        must_revalidate,
    )


DEFAULT_MEDIA_TYPES = {
    "array": {"*/*": "application/octet-stream", "image/*": "image/png"},
    "dataframe": {"*/*": APACHE_ARROW_FILE_MIME_TYPE},
    "node": {"*/*": "application/x-hdf5"},
    "xarray_data_array": {"*/*": "application/octet-stream"},
    "xarray_dataset": {"*/*": "application/netcdf"},
}


def construct_data_response(
    structure_family,
    serialization_registry,
    payload,
    metadata,
    request,
    format=None,
    specs=None,
    expires=None,
    filename=None,
):
    request.state.endpoint = "data"
    if specs is None:
        specs = []
    default_media_type = DEFAULT_MEDIA_TYPES[structure_family]["*/*"]
    # Give priority to the `format` query parameter. Otherwise, consult Accept
    # header.
    if format is not None:
        media_types_or_aliases = format.split(",")
        # Resolve aliases, like "csv" -> "text/csv".
        media_types = [
            serialization_registry.resolve_alias(t) for t in media_types_or_aliases
        ]
    else:
        # The HTTP spec says these should be separated by ", " but some
        # browsers separate with just "," (no space).
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation/List_of_default_Accept_values#default_values  # noqa
        # That variation is what we are handling below with lstrip.
        media_types = [
            s.lstrip(" ")
            for s in request.headers.get("Accept", default_media_type).split(",")
        ]

    # The client may give us a choice of media types. Find the first one
    # that we support.
    supported = set()
    for media_type in media_types:
        if media_type == "*/*":
            media_type = DEFAULT_MEDIA_TYPES[structure_family]["*/*"]
        elif structure_family == "array" and media_type == "image/*":
            media_type = DEFAULT_MEDIA_TYPES[structure_family]["image/*"]
        # fall back to generic dataframe serializer if no specs present
        for spec in specs + [structure_family]:
            media_types_for_spec = serialization_registry.media_types(spec)
            if media_type in media_types_for_spec:
                break
            supported.update(media_types_for_spec)
        else:
            # None of the specs or the structure_family can serialize to this
            # media_type. Try the next one.
            continue
        # We found a match above. We have our media_type.
        break
    else:
        # We have checked each of the media_types, and we cannot serialize
        # to any of them.
        raise UnsupportedMediaTypes(
            f"None of the media types requested by the client are supported. "
            f"Supported: {', '.join(supported)}. Requested: {', '.join(media_types)}.",
        )
    with record_timing(request.state.metrics, "tok"):
        # Create an ETag that uniquely identifies this content and the media
        # type that it will be encoded as.
        etag = tokenize((payload, media_type))
    headers = {"ETag": etag}
    if expires is not None:
        headers["Expires"] = expires.strftime(HTTP_EXPIRES_HEADER_FORMAT)
    if request.headers.get("If-None-Match", "") == etag:
        # If the client already has this content, confirm that.
        return Response(status_code=304, headers=headers)
    if filename:
        headers["Content-Disposition"] = f"attachment;filename={filename}"
    # This is the expensive step: actually serialize.
    try:
        content = serialization_registry(
            structure_family, media_type, payload, metadata
        )
    except UnsupportedShape as err:
        raise UnsupportedMediaTypes(
            f"The shape of this data {err.args[0]} is incompatible with the requested format ({media_type}). "
            f"Slice it or choose a different format.",
        )
    except SerializationError as err:
        raise UnsupportedMediaTypes(
            f"This type is supported in general but there was an error packing this specific data: {err.args[0]}",
        )
    return PatchedResponse(
        content=content,
        media_type=media_type,
        headers=headers,
    )


def construct_resource(
    base_url,
    path_parts,
    entry,
    fields,
    select_metadata,
    omit_links,
    media_type,
):
    path_str = "/".join(path_parts)
    attributes = {"ancestors": path_parts[:-1]}
    if schemas.EntryFields.metadata in fields:
        if select_metadata is not None:
            attributes["metadata"] = jmespath.compile(select_metadata).search(
                entry.metadata
            )
        else:
            attributes["metadata"] = entry.metadata
    if schemas.EntryFields.specs in fields:
        attributes["specs"] = getattr(entry, "specs", [])
    if (entry is not None) and entry.structure_family == "node":
        attributes["structure_family"] = "node"
        if schemas.EntryFields.count in fields:
            attributes["count"] = len_or_approx(entry)
            if hasattr(entry, "sorting"):
                # In the Python API we encode sorting as (key, direction).
                # This order-based "record" notion does not play well with OpenAPI.
                # In the HTTP API, therefore, we use {"key": key, "direction": direction}.
                attributes["sorting"] = [
                    {"key": key, "direction": direction}
                    for key, direction in entry.sorting
                ]
        d = {
            "id": path_parts[-1] if path_parts else "",
            "attributes": schemas.NodeAttributes(**attributes),
        }
        if not omit_links:
            d["links"] = {
                "self": f"{base_url}/node/metadata/{path_str}",
                "search": f"{base_url}/node/search/{path_str}",
                "full": f"{base_url}/node/full/{path_str}",
            }
        resource = schemas.Resource[
            schemas.NodeAttributes, schemas.NodeLinks, schemas.NodeMeta
        ](**d)
    else:
        links = {"self": f"{base_url}/node/metadata/{path_str}"}
        structure = {}
        if entry is not None:
            # entry is None when we are pulling just *keys* from the
            # Tree and not values.
            ResourceLinksT = schemas.resource_links_type_by_structure_family[
                entry.structure_family
            ]
            links.update(
                {
                    link: template.format(base_url=base_url, path=path_str)
                    for link, template in FULL_LINKS[entry.structure_family].items()
                }
            )
            if schemas.EntryFields.structure_family in fields:
                attributes["structure_family"] = entry.structure_family
            if schemas.EntryFields.macrostructure in fields:
                macrostructure = entry.macrostructure()
                if macrostructure is not None:
                    structure["macro"] = dataclasses.asdict(macrostructure)
            if schemas.EntryFields.microstructure in fields:
                if entry.structure_family == "node":
                    assert False  # not sure if this ever happens
                    pass
                elif entry.structure_family == "dataframe":
                    import pandas

                    microstructure = entry.microstructure()
                    arrow_encoded_meta = bytes(serialize_arrow(microstructure.meta, {}))
                    divisions_wrapped_in_df = pandas.DataFrame(
                        {"divisions": list(microstructure.divisions)}
                    )
                    arrow_encoded_divisions = bytes(
                        serialize_arrow(divisions_wrapped_in_df, {})
                    )
                    if media_type == "application/json":
                        # For JSON, base64-encode the binary Arrow-encoded data,
                        # and indicate that this has been done in the data URI.
                        data_uri = f"data:{APACHE_ARROW_FILE_MIME_TYPE};base64,"
                        arrow_encoded_meta = (
                            data_uri + base64.b64encode(arrow_encoded_meta).decode()
                        )
                        arrow_encoded_divisions = (
                            data_uri
                            + base64.b64encode(arrow_encoded_divisions).decode()
                        )
                    else:
                        # In msgpack, we can encode the binary Arrow-encoded data directly.
                        assert media_type == "application/x-msgpack"
                    structure["micro"] = {
                        "meta": arrow_encoded_meta,
                        "divisions": arrow_encoded_divisions,
                    }
                else:
                    microstructure = entry.microstructure()
                    if microstructure is not None:
                        structure["micro"] = dataclasses.asdict(microstructure)
            if entry.structure_family == "array":
                shape = structure.get("macro", {}).get("shape")
                if shape is None:
                    # The client did not request structure so we have not yet
                    # accessed it, and we have access it specifically to construct this link.
                    shape = entry.macrostructure().shape
                block_template = ",".join(
                    f"{{index_{index}}}" for index in range(len(shape))
                )
                links[
                    "block"
                ] = f"{base_url}/array/block/{path_str}?block={block_template}"
            elif entry.structure_family == "dataframe":
                links[
                    "partition"
                ] = f"{base_url}/dataframe/partition/{path_str}?partition={{index}}"
            attributes["structure"] = structure
        else:
            # We only have entry names, not structure_family, so
            ResourceLinksT = schemas.SelfLinkOnly
        d = {
            "id": path_parts[-1],
            "attributes": schemas.NodeAttributes(**attributes),
        }
        if not omit_links:
            d["links"] = links
        resource = schemas.Resource[
            schemas.NodeAttributes, ResourceLinksT, schemas.EmptyDict
        ](**d)
    return resource


class PatchedResponse(Response):
    "Patch the render method to accept memoryview."

    def render(self, content: Any) -> bytes:
        if isinstance(content, memoryview):
            return content.cast("B")
        return super().render(content)


class PatchedStreamingResponse(StreamingResponse):
    "Patch the stream_response method to accept memoryview."

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            # BEGIN ALTERATION
            if not isinstance(chunk, (bytes, memoryview)):
                # END ALTERATION
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})


class NumpySafeJSONResponse(JSONResponse):
    def __init__(self, *args, metrics, **kwargs):
        self.__metrics = metrics
        super().__init__(*args, **kwargs)

    def render(self, content: Any) -> bytes:
        with record_timing(self.__metrics, "pack"):
            return safe_json_dump(content)


def _fallback_msgpack_encoder(obj):
    # If numpy has not been imported yet, then we can be sure that obj
    # is not a numpy object, and we want to avoid triggering a numpy
    # import. (The server does not have a hard numpy dependency.)
    if "numpy" in sys.modules:
        import numpy

        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            if numpy.isscalar(obj):
                return obj.item()
            return obj.tolist()
    if isinstance(obj, uuid.UUID):
        return str(obj)  # hyphen-separated hex per RFC4122
    return obj


def _patch_naive_datetimes(obj):
    """
    If a naive datetime is found, attach local time.

    Msgpack can only serialize datetimes with tzinfo.
    """
    if hasattr(obj, "items"):
        patched_obj = {}
        for k, v in obj.items():
            patched_obj[k] = _patch_naive_datetimes(v)
    elif (not isinstance(obj, str)) and isinstance(obj, collections.abc.Iterable):
        patched_obj = []
        for item in obj:
            patched_obj.append(_patch_naive_datetimes(item))
    elif isinstance(obj, datetime) and obj.tzinfo is None:
        patched_obj = obj.astimezone(_LOCAL_TZINFO)
    else:
        patched_obj = obj
    return patched_obj


class MsgpackResponse(Response):
    media_type = "application/x-msgpack"

    def __init__(self, *args, metrics, **kwargs):
        self.__metrics = metrics
        super().__init__(*args, **kwargs)

    def render(self, content: Any, _reentered=False) -> bytes:
        try:
            with record_timing(self.__metrics, "pack"):
                return msgpack.packb(
                    content, default=_fallback_msgpack_encoder, datetime=True
                )
        except (ValueError, TypeError) as err:
            # msgpack tries to handle all datetimes, but if it
            # received a naive one (tzinfo=None) then it fails.
            # We cannot use the default hook to handle this because
            # it is not called.
            if "can not serialize 'datetime.datetime' object" in str(err) and (
                not _reentered
            ):
                patched_content = _patch_naive_datetimes(content)
                return self.render(patched_content, _reentered=True)
            raise


JSON_MIME_TYPE = "application/json"
MSGPACK_MIME_TYPE = "application/x-msgpack"
# This is a silly time format, but it is the HTTP standard.
HTTP_EXPIRES_HEADER_FORMAT = "%a, %d %b %Y %H:%M:%S GMT"


def resolve_media_type(request):
    media_types = request.headers.get("Accept", JSON_MIME_TYPE).split(", ")
    for media_type in media_types:
        if media_type == "*/*":
            media_type = JSON_MIME_TYPE
            break
        if media_type == MSGPACK_MIME_TYPE:
            break
        if media_type == JSON_MIME_TYPE:
            break
    else:
        # It is commmon in HTTP to fall back on a default representation if
        # none of the requested ones are available. We do not do this for
        # data payloads, but it makes some sense to do it for these metadata
        # messages.
        media_type = JSON_MIME_TYPE
    assert media_type in {JSON_MIME_TYPE, MSGPACK_MIME_TYPE}
    return media_type


def json_or_msgpack(request, content, expires=None, headers=None):
    media_type = resolve_media_type(request)
    content_as_dict = content.dict()
    with record_timing(request.state.metrics, "tok"):
        etag = md5(str(content_as_dict).encode()).hexdigest()
    headers = headers or {}
    headers["ETag"] = etag
    if expires is not None:
        headers["Expires"] = expires.strftime(HTTP_EXPIRES_HEADER_FORMAT)
    if request.headers.get("If-None-Match", "") == etag:
        # If the client already has this content, confirm that.
        return Response(status_code=304, headers=headers)
    if media_type == "application/x-msgpack":
        return MsgpackResponse(
            content_as_dict, headers=headers, metrics=request.state.metrics
        )
    return NumpySafeJSONResponse(
        content_as_dict, headers=headers, metrics=request.state.metrics
    )


class UnsupportedMediaTypes(Exception):
    pass


class NoEntry(KeyError):
    pass


class WrongTypeForRoute(Exception):
    pass


FULL_LINKS = {
    "node": {"full": "{base_url}/node/full/{path}"},
    "array": {"full": "{base_url}/array/full/{path}"},
    "dataframe": {"full": "{base_url}/node/full/{path}"},
    "xarray_data_array": {
        "full_variable": "{base_url}/array/full/{path}/variable",
    },
    "xarray_dataset": {
        "full_variable": "{base_url}/array/full/{path}/data_vars/{{variable}}/variable",
        "full_coord": "{base_url}/array/full/{path}/coords/{{coord}}/variable",
        "full_dataset": "{base_url}/node/full/{path}",
    },
}



from pathlib import Path
from typing import List, Optional

import typer

cli_app = typer.Typer()
serve_app = typer.Typer()
profile_app = typer.Typer()
api_key_app = typer.Typer()
admin_app = typer.Typer()
cli_app.add_typer(serve_app, name="serve", help="Launch a Tiled server.")
cli_app.add_typer(
    profile_app, name="profile", help="Examine Tiled 'profiles' (client-side config)."
)
cli_app.add_typer(
    api_key_app, name="api_key", help="Create, list, and revoke API keys."
)
cli_app.add_typer(
    admin_app,
    name="admin",
    help="Administrative utilities for managing large deployments.",
)


@admin_app.command("initialize-database")
def initialize_database(database_uri: str):
    from sqlalchemy import create_engine

    from ..database.core import (
        REQUIRED_REVISION,
        UninitializedDatabase,
        check_database,
        initialize_database,
    )

    engine = create_engine(database_uri)
    redacted_url = engine.url._replace(password="[redacted]")
    try:
        check_database(engine)
    except UninitializedDatabase:
        # Create tables and stamp (alembic) revision.
        typer.echo(
            f"Database {redacted_url} is new. Creating tables and marking revision {REQUIRED_REVISION}."
        )
        initialize_database(engine)
        typer.echo("Database initialized.")
    else:
        typer.echo(f"Database at {redacted_url} is already initialized.")
        raise typer.Abort()


@cli_app.command("login")
def login(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
    show_secret_tokens: bool = typer.Option(
        False, "--show-secret-tokens", help="Show secret tokens after successful login."
    ),
    refresh_or_fail: bool = typer.Option(
        False, "--refresh-or-fail", help="Refresh authentication or exit with failure."
    ),
):
    """
    Log in to an authenticated Tiled server.
    """
    from tiled.client.utils import ClientError

    options = {}
    if refresh_or_fail:
        options["prompt_for_reauthentication"] = "never"
    try:
        client = _client_from_uri_or_profile(
            uri_or_profile, no_verify=no_verify, **options
        )
    except (ValueError, ClientError) as err:
        (msg,) = err.args
        typer.echo(msg)
        raise typer.Abort()
    if show_secret_tokens:
        from pprint import pformat

        typer.echo(pformat(dict(client.context.tokens)))


@cli_app.command("sessions")
def sessions(
    show_secret_tokens: bool = typer.Option(
        False, "--show-secret-tokens", help="Show secret (refresh) tokens."
    ),
):
    """
    List all authenticated Tiled sessions.
    """
    from tiled.client.context import sessions

    if show_secret_tokens:
        sessions_ = sessions()
        max_netloc_len = max(len(netloc) for netloc in sessions_)
        for netloc, token in sessions_.items():
            padding = max_netloc_len - len(netloc)
            typer.echo(f"{netloc}{' ' * padding}   {token}")
    else:
        for netloc in sessions():
            typer.echo(netloc)


@api_key_app.command("create")
def create_api_key(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    expires_in: Optional[int] = typer.Option(
        None,
        help=(
            "Number of seconds until API key expires. If None, "
            "it will never expire or it will have the maximum lifetime "
            "allowed by the server."
        ),
    ),
    scopes: Optional[List[str]] = typer.Option(
        None,
        help=(
            "Restrict the access available to this API key by listing scopes. "
            "By default, it will inherit the scopes of its owner."
        ),
    ),
    note: Optional[str] = typer.Option(None, help="Add a note to label this API key."),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    client = _client_from_uri_or_profile(uri_or_profile, no_verify=no_verify)
    if not scopes:
        # This is how typer interprets unspecified scopes.
        # Replace with None to get default scopes.
        scopes = None
    info = client.context.create_api_key(
        scopes=scopes, expires_in=expires_in, note=note
    )
    # TODO Print other info to the stderr?
    typer.echo(info["secret"])


@api_key_app.command("list")
def list_api_keys(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    client = _client_from_uri_or_profile(uri_or_profile, no_verify=no_verify)
    info = client.context.whoami()
    if not info["api_keys"]:
        type.echo("No API keys found")
        return
    max_note_len = max(len(api_key["note"] or "") for api_key in info["api_keys"])
    COLUMNS = f"First 8   Expires at (UTC)     Latest activity      Note{' ' * (max_note_len - 4)}  Scopes"
    typer.echo(COLUMNS)
    for api_key in info["api_keys"]:
        note_padding = 2 + max_note_len - len(api_key["note"] or "")
        if api_key["expiration_time"] is None:
            expiration_time = "-"
        else:
            expiration_time = (
                api_key["expiration_time"]
                .replace(microsecond=0, tzinfo=None)
                .isoformat()
            )
        if api_key["latest_activity"] is None:
            latest_activity = "-"
        else:
            latest_activity = (
                api_key["latest_activity"]
                .replace(microsecond=0, tzinfo=None)
                .isoformat()
            )
        typer.echo(
            (
                f"{api_key['first_eight']:10}"
                f"{expiration_time:21}"
                f"{latest_activity:21}"
                f"{(api_key['note'] or '')}{' ' * note_padding}"
                f"{' '.join(api_key['scopes']) or '-'}"
            )
        )


@api_key_app.command("revoke")
def revoke_api_key(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    first_eight: str = typer.Argument(
        ..., help="First eight characters of API key (or the whole key)"
    ),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    client = _client_from_uri_or_profile(uri_or_profile, no_verify=no_verify)
    client.context.revoke_api_key(first_eight[:8])


@cli_app.command("logout")
def logout(
    uri_or_profile: Optional[str] = typer.Argument(
        None, help="URI 'http[s]://...' or a profile name. If blank, log out of all."
    ),
):
    """
    Log out from one or all authenticated Tiled servers.
    """
    from tiled.client.context import logout, logout_all

    logged_out_from = []
    if uri_or_profile is None:
        logged_out_from.extend(logout_all())
    else:
        logged_out_from.append(logout(uri_or_profile))
    for netloc in logged_out_from:
        typer.echo(netloc)


@cli_app.command("tree")
def tree(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    max_lines: int = typer.Argument(20, help="Max lines to show."),
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    """
    Show the names of entries in a Tree.

    This is similar to the UNIX utility `tree` for listing nested directories.
    """
    from ..utils import gen_tree

    tree_obj = _client_from_uri_or_profile(uri_or_profile, no_verify=no_verify)
    for counter, line in enumerate(gen_tree(tree_obj), start=1):
        if (max_lines is not None) and (counter > max_lines):
            print(
                f"Output truncated at {max_lines} lines. "
                "Use `tiled tree CATALOG <N>` to see <N> lines."
            )
            break
        print(line)


@cli_app.command("download")
def download(
    uri_or_profile: str = typer.Argument(
        ..., help="URI 'http[s]://...' or a profile name."
    ),
    cache_path: str = typer.Argument(..., help="Local directory for cache storage"),
    available_bytes: Optional[int] = None,
    no_verify: bool = typer.Option(False, "--no-verify", help="Skip SSL verification."),
):
    """
    Download content from a Tree to an on-disk cache.
    """
    from ..client.cache import Cache, download

    cache = Cache.on_disk(cache_path, available_bytes=available_bytes)
    client = _client_from_uri_or_profile(
        uri_or_profile, cache=cache, no_verify=no_verify
    )
    download(client)


@profile_app.command("paths")
def profile_paths():
    "List the locations that the client will search for profiles (client-side configuration)."
    from ..profiles import paths

    print("\n".join(str(p) for p in paths))


@profile_app.command("list")
def profile_list():
    "List the profiles (client-side configuration) found and the files they were read from."
    from ..profiles import load_profiles

    profiles = load_profiles()
    if not profiles:
        typer.echo("No profiles found.")
        return
    max_len = max(len(name) for name in profiles)
    PADDING = 4

    print(
        "\n".join(
            f"{name:<{max_len + PADDING}}{filepath}"
            for name, (filepath, _) in profiles.items()
        )
    )


@profile_app.command("show")
def profile_show(profile_name: str):
    "Show the content of a profile."
    import sys

    import yaml

    from ..profiles import load_profiles

    profiles = load_profiles()
    try:
        filepath, content = profiles[profile_name]
    except KeyError:
        typer.echo(
            f"The profile {profile_name!r} could not be found. "
            "Use tiled profile list to see profile names."
        )
        raise typer.Abort()
    print(f"Source: {filepath}", file=sys.stderr)
    print("--", file=sys.stderr)
    print(yaml.dump(content), file=sys.stdout)


@serve_app.command("directory")
def serve_directory(
    directory: str,
    public: bool = typer.Option(False, "--public"),
    keep_ext: bool = typer.Option(
        False,
        "--keep-ext",
        help=(
            "Serve a file like 'measurements.csv' as its full filepath with extension, "
            "instead of the default which would serve it as 'measurements'. "
            "This is discouraged because it leaks details about the storage "
            "format to the client, such that changing the storage in the future "
            "may break user (client-side) code."
        ),
    ),
    poll_interval: float = typer.Option(
        None,
        "--poll-interval",
        help=(
            "Time in seconds between scans of the directory for removed or "
            "changed files. If 0, do not poll for changes."
        ),
    ),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
):
    "Serve a Tree instance from a directory of files."
    from ..adapters.files import DirectoryAdapter
    from ..server.app import build_app, print_admin_api_key_if_generated

    tree_kwargs = {}
    server_settings = {}
    if keep_ext:
        from ..adapters.files import identity

        tree_kwargs.update({"key_from_filename": identity})
    if poll_interval is not None:
        tree_kwargs.update({"poll_interval": poll_interval})
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    tree = DirectoryAdapter.from_directory(directory, **tree_kwargs)
    web_app = build_app(tree, {"allow_anonymous_access": public}, server_settings)
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("pyobject")
def serve_pyobject(
    object_path: str = typer.Argument(
        ..., help="Object path, as in 'package.subpackage.module:object_name'"
    ),
    public: bool = typer.Option(False, "--public"),
    host: str = typer.Option(
        "127.0.0.1",
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`."
        ),
    ),
    port: int = typer.Option(8000, help="Bind to a socket with this port."),
    object_cache_available_bytes: Optional[float] = typer.Option(
        None,
        "--data-cache",
        help=(
            "Maximum size for the object cache, given as a number of bytes as in "
            "1_000_000 or as a fraction of system RAM (total physical memory) as in "
            "0.3. Set to 0 to disable this cache. By default, it will use up to "
            "0.15 (15%) of RAM."
        ),
    ),
):
    "Serve a Tree instance from a Python module."
    from ..server.app import build_app, print_admin_api_key_if_generated
    from ..utils import import_object

    tree = import_object(object_path)
    server_settings = {}
    if object_cache_available_bytes is not None:
        server_settings["object_cache"] = {}
        server_settings["object_cache"][
            "available_bytes"
        ] = object_cache_available_bytes
    web_app = build_app(tree, {"allow_anonymous_access": public}, server_settings)
    print_admin_api_key_if_generated(web_app, host=host, port=port)

    import uvicorn

    uvicorn.run(web_app, host=host, port=port)


@serve_app.command("config")
def serve_config(
    config_path: Path = typer.Argument(
        None,
        help=(
            "Path to a config file or directory of config files. "
            "If None, check environment variable TILED_CONFIG. "
            "If that is unset, try default location ./config.yml."
        ),
    ),
    public: bool = typer.Option(False, "--public"),
    host: str = typer.Option(
        None,
        help=(
            "Bind socket to this host. Use `--host 0.0.0.0` to make the application "
            "available on your local network. IPv6 addresses are supported, for "
            "example: --host `'::'`. Uses value in config by default."
        ),
    ),
    port: int = typer.Option(
        None, help="Bind to a socket with this port. Uses value in config by default."
    ),
):
    "Serve a Tree as specified in configuration file(s)."
    import os

    from ..config import construct_build_app_kwargs, parse_configs

    config_path = config_path or os.getenv("TILED_CONFIG", "config.yml")
    try:
        parsed_config = parse_configs(config_path)
    except Exception as err:
        typer.echo(str(err))
        raise typer.Abort()

    # Let --public flag override config.
    if public:
        if "authentication" not in parsed_config:
            parsed_config["authentication"] = {}
        parsed_config["authentication"]["allow_anonymous_access"] = True

    # Delay this import so that we can fail faster if config-parsing fails above.

    from ..server.app import build_app, logger, print_admin_api_key_if_generated

    # Extract config for uvicorn.
    uvicorn_kwargs = parsed_config.pop("uvicorn", {})
    # If --host is given, it overrides host in config. Same for --port.
    uvicorn_kwargs["host"] = host or uvicorn_kwargs.get("host", "127.0.0.1")
    uvicorn_kwargs["port"] = port or uvicorn_kwargs.get("port", 8000)

    # This config was already validated when it was parsed. Do not re-validate.
    kwargs = construct_build_app_kwargs(parsed_config, source_filepath=config_path)
    logger.info(f"Using configuration from {Path(config_path).absolute()}")
    web_app = build_app(**kwargs)
    print_admin_api_key_if_generated(
        web_app, host=uvicorn_kwargs["host"], port=uvicorn_kwargs["port"]
    )

    # Likewise, delay this import.

    import uvicorn

    uvicorn.run(web_app, **uvicorn_kwargs)


def _client_from_uri_or_profile(
    uri_or_profile, no_verify, cache=None, prompt_for_reauthentication=None
):
    from ..client import from_profile, from_uri

    options = {}
    if no_verify:
        options["verify"] = False
    if prompt_for_reauthentication is not None:
        options["prompt_for_reauthentication"] = prompt_for_reauthentication
    if uri_or_profile.startswith("http://") or uri_or_profile.startswith("https://"):
        # This looks like a URI.
        uri = uri_or_profile
        return from_uri(uri, cache=cache, **options)
    else:
        from ..profiles import list_profiles

        # Is this a profile name?
        if uri_or_profile in list_profiles():
            profile_name = uri_or_profile
            return from_profile(profile_name, cache=cache, **options)
        typer.echo(
            f"Not sure what to do with tree {uri_or_profile!r}. "
            "It does not look like a URI (it does not start with http[s]://) "
            "and it does not match any profiles. Use `tiled profiles list` to "
            "see profiles."
        )
        raise typer.Abort()


main = cli_app

if __name__ == "__main__":
    main()

# This object is used by the auto-generated documentation.
typer_click_object = typer.main.get_command(cli_app)
